import subprocess
import os

from hydenv.util import env
from hydenv.exercises import HydenvExercises

SCRIPTPATH = os.path.join(os.path.dirname(__file__), 'main.py')


def run(connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}", legacy=False, **kwargs):
    """
    """
    connection = env.build_connection(connection=connection)
    # check if the legacy exercises app should be used
    if legacy:
        backend = kwargs.get('backend', 'local')
        port = kwargs.get('port', 5000)
        exercisesApi = HydenvExercises(connection=connection, backend=backend, port=port)
        exercisesApi.start(no_browser=kwargs.get('no_browser', False))
        return

    # run the new GUI
    os.system(f"streamlit run {SCRIPTPATH}")



import os
import types
import inspect
import json

from sqlalchemy.orm.session import sessionmaker
from sqlalchemy import create_engine

from hydenv import models
from hydenv.util import env, check_overrides, stringify_output


class HydenvMeasurements:
    """
    Measurements management tool\n
    This is the main tool for CRUD operations on the database.
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>
    """
    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}"):
        # substitute and save
        self.__connection = env.build_connection(connection=connection)
        self.__SessionCls = sessionmaker(bind=create_engine(self.__connection))
        self.session = self.__SessionCls()

    def create_metadata(self, device_id=None, location=None, description=None, **kwargs):
        """
        Create a new metadata record.\n
        :param device_id: unique identifier for the device used
        :param location: The device location. can be tuple, string
        :param description: A short description of the campaign.
        """
        if isinstance(location, (list, tuple)):
            geom = 'SRID=4326;POINT ({x} {y})'.format(x=location[0], y=location[1])
        if not isinstance(location, str):
            raise AttributeError('location has to be of type list (x,y) or string (WKT).')
        else:
            geom = location
        if not geom.startswith('SRID=4326;'):
            geom = 'SRID=4326;%s' % geom
        if not 'POINT' in geom:
            raise AttributeError('location has to be of shape Point')

        args = dict(device_id=device_id, location=geom, description=description, **kwargs)
        
        return self.create('Metadata', skip_overrides=True, **args)

    @check_overrides
    def create(self, model_name, **kwargs):
        """
        Create a new record in the database\n
        This is the most generic function, that does not handle 
        specific database entities.
        :param model_name: name of the model entity
        :param skip_overrides: If True, override functions will not be called
        """        
        # get the model class
        Model = getattr(models, model_name)
        model = Model(**kwargs)

        try:
            self.session.add(model)
            self.session.commit()
        except Exception as e:
            self.session.rollback()
            raise e

        return model

    @check_overrides
    def read(self, model_name, return_query=False, **kwargs):
        """
        Read a database record\n
        The data will be returned as a list of model instances.
        :param model_name: name of the model entity
        :param skip_overrides: If True, override functions will not be called
        :param return_query: If True, the SQL query instead of the result\n
            will be returned.
        """
        Model = getattr(models, model_name)
        columns = Model.__table__.columns

        # build the query
        query = self.session.query(Model)

        # add the kwargs
        for name, value in kwargs.items():
            if name in columns:
                # exact match
                query = query.filter(columns.get(name)==value)
        
        # return
        if return_query:
            return query
        else:
            return query.all()

    @check_overrides
    def update(self, model_name, id, **kwargs):
        """
        Update an existing model.\n
        :param model_name: name of the model entity
        :param skip_overrides: If True, override functions will not be called
        :param id: The id of the updated object
        """
        # this will raise an exception if not exactly one instance is found
        instance = HydenvMeasurements.read(self, model_name, return_query=True, id=id).one()
        print(instance)
        
        # update
        for name, value in kwargs.items():
            setattr(instance, name, value)
        
        try:
            self.session.add(instance)
            self.session.commit()
        except Exception as e:
            self.session.rollback()
            raise e

        return instance

    @check_overrides
    def delete(self, model_name, id=None, **kwargs):
        """
        Delete an existing record.\n
        You can either delete by id, to exactly control which records 
        are deleted, or you can pass in any kind of filter statement
        to delete all matching instances. In this case you have to 
        omit the id parameter.
        Caution: if you omit the id, but do not pass a filter condition,
        **all** records will be deleted.
        :param model_name: name of the model entity
        :param skip_overrides: If True, override functions will not be called
        :param id: The id of the updated object
        """
        if id is not None:
            instance = HydenvMeasurements.read(self, model_name, return_query=True, id=id)
            print(instance)
            instances = [instance.one()]
        else:
            instances = HydenvMeasurements.read(self, model_name, **kwargs)
        
        # delete
        try:
            for inst in instances:
                self.session.delete(inst)
            self.session.commit()
        except Exception as e:
            self.session.rollback()
            raise e

        print('Deleted %d %s instances' % (len(instances), model_name))

class HydenvMeasurementsCli(HydenvMeasurements):
    def __init__(self, fmt='json', connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}"):
        """
        Measurements management tool\n
        This is the main tool for CRUD operations on the database.
        :param connection: The database URI following syntax:\n
            postgresql://<user>:<password>@<host>:<port>/<database>
        :param fmt: output format - can be one of:\n
            json    -> JSON output (default)
            print   -> raw print to StdOut
            csv     -> not supported yet
        """
        super(HydenvMeasurementsCli, self).__init__(connection)
        self.fmt = fmt

# attach the stringify decorator to all CRUD function
for name, fn in inspect.getmembers(HydenvMeasurementsCli):
    if isinstance(fn, types.FunctionType) and any([_ in name for _ in ('create', 'read', 'update', 'delete')]):
        setattr(HydenvMeasurementsCli, name, stringify_output(fn))


if __name__=='__main__':
    import fire
    fire.Fire(HydenvMeasurementsCli)



import os
import glob

import geopandas as gpd
import pandas as pd
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from tqdm import tqdm

from hydenv.util import env
from hydenv import models


class HydenvCamelsDeExample:
    """
    CAMELS-DE Dataset Example Loader.\n
    Loads the german CAMELS-DE dataset to the database. Please beaware, that the CAMELS-DE dataset
    is not yet published and you need a copy of the processing files on your drive, as this API does
    as of now only work with local files.
    The database has to be installed and initialized before. 
    Do that by hand or use the two CLI commands:
        python -m hydenv database install  --connection=postgresql://postgres:<adminpassword>@localhost:5432/postgres
        python -m hydenv database init --clean --connection=postgresql://hydenv:hydenv@localhost:5432/hydenv
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>
    :param source_dir: string - overwrite the default data url to load other data. 
    :param quiet: bool - suppresses print output to stdout
    :param only: string - load --only=metadata for omitting the data upload. If not set, you need to specify the source_dir 
    :param nuts-filter: Filter for only one or more federal state by using its NUTS lvl2 ID (ie. DE1); use 'all' to load all

    """
    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}", source_dir: str = None, quiet: bool = True): 
        # set the source param
        self.source_dir = source_dir
        self.quiet = quiet

        # create a database session
        self.__connection = env.build_connection(connection=connection)
        self.engine = create_engine(self.__connection)
#        Session = sessionmaker(bind=self.engine)
#        self.session = Session()

        # set the URL to the camels-de api
        self._state_api_url = "https://api.camels-de.org/state"

    def run(self, only: str = None, nuts_filter: str = 'all'):
        # TODO implement this, when the data is published.
        if self.source_dir is None and only != 'metadata':
            raise NotImplementedError("As of now, the CAMELS-DE data itself is not published. You need a local copy and reference it using '--source-dir=/path/to/output_data'")

        # run each importer function
        # metadata
        if only is None or only == 'metadata':
            self._upload_metadata(nuts_filter=nuts_filter)
        
        # data
        if only is None or only == 'data':
            self._upload_data(nuts_filter=nuts_filter)
    
    def _upload_metadata(self, nuts_filter: str):
        # checkout the filter
        if nuts_filter == 'all' or nuts_filter is None:
            url = f"{self._state_api_url}/metadata.csv"
        else:
            url = f"{self._state_api_url}/{nuts_filter}/metadata.csv"

        # download the metadata
        if not self.quiet:
            print(f'Downloading CAMELS-DE metadata from: {url}...', end='')
        df = pd.read_csv(url)
        gdf = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df.x, df.y), crs=3035)
        if not self.quiet:
            print('done.')

        # uploading the metadata
        if not self.quiet:
            print(f'Uploading to {str(self.engine)}...', end='')
        gdf.to_postgis('camels_de_metadata', con=self.engine, if_exists='append')
        if not self.quiet:
            print('done.')
    
    def _upload_data(self, nuts_filter: str):
        # Right now, complete uploads are not accepted
        if nuts_filter is None:
            raise NotImplementedError("As of now, you can't upload CAMELS-DE data without federal state filter. Add a filter or pass --only=metadata flag.")
        
        # get the path
        base_path = os.path.join(self.source_dir, nuts_filter.upper())
        if not os.path.exists(base_path):
            raise OSError(f"The path {base_path} does not exist")
        
        # get all files
        files = glob.glob(os.path.join(base_path, '*', '*_data.csv'))
        if not self.quiet:
            print(f'Found {len(files)} files. Start uploading...')
        
        # upload
        for fname in tqdm(files, unit=' files'):
            # read the file
            df = pd.read_csv(fname)

            # add the id
            camels_id = os.path.basename(fname).split('_')[0]
            df['camels_id'] = camels_id

            # upload
            df.to_sql('camels_de_data', self.engine, if_exists='append')
        
        if not self.quiet:
            print('\nDone!')

    def drop(self):
        # drop the tables if they exist
        with self.engine.connect() as con:
            con.execute("DROP TABLE IF EXISTS camels_de_metadata;")
            con.execute("DROP TABLE IF EXISTS camels_de_data;")
            con.execute("COMMIT;")


from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, ForeignKey, UniqueConstraint
from sqlalchemy import Integer, BigInteger, String, DateTime, Numeric, Boolean, REAL
from sqlalchemy.orm import relationship, Session, object_session
from geoalchemy2 import Geometry

from datetime import datetime as dt


class Base(object):
    # overwrite the default values
    __prepopulate__ = []

    @classmethod
    def defaults(cls, session: Session):
        instances = [cls(**d) for d in cls.__prepopulate__]

        try:
            session.add_all(instances)
            session.commit()
        except Exception as e:
            session.rollback()
            print('Populating defaults errored\n%s' % str(e))

# augment the sqlalchemy Base class with defaults classmethod
Base = declarative_base(cls=Base)

# ---------------------------------------------------------------------
# Database Models
# ---------------------------------------------------------------------
class DetailAssociation(Base):
    __tablename__ = 'nm_metadata_details'

    meta_id = Column(Integer, ForeignKey('metadata.id'), primary_key=True)
    detail_id = Column(Integer, ForeignKey('details.id'), primary_key=True)

class Metadata(Base):
    __tablename__ = 'metadata'

    # columns
    id = Column(Integer, primary_key=True)
    device_id = Column(String, nullable=False)
    sensor_id = Column(Integer, ForeignKey('sensors.id'))
    term_id = Column(Integer, ForeignKey('terms.id'))
    location = Column(Geometry('POINT', srid=4326), nullable=True)
    description = Column(String, nullable=True)

    # relationships
    term = relationship('Term', back_populates='data')
    details = relationship('Detail', secondary='nm_metadata_details', back_populates='meta', cascade='all,delete')
    sensor  = relationship('Sensor', back_populates='meta')
    raw_data = relationship('RawData', back_populates='meta', cascade='all,delete')
    data = relationship('Data', back_populates='meta', cascade='all,delete')

    def __init__(self, session=None, **kwargs):
        # extract the column names of this model
        column_names = [col.name for col in Metadata.__table__.columns]

        # split the input into core and details
        in_names = {k: v for k, v in kwargs.items() if k in column_names}
        out_names = {k: v for k, v in kwargs.items() if k not in column_names}
        
        # run Base class __init__ method
        super(Metadata, self).__init__(**in_names)

        # add details
        self.set_details(out_names, session)

    def get_details(self):
        details = dict()
        for detail in self.details:
            detail.to_dict(add_to=details)
        return details

    def set_details(self, details: dict, session=None):
        # append all details
        for k,v in details.items():
            self.details.append(Detail.get_or_init(session, k, v))


class Variable(Base):
    __tablename__ = 'variables'
    __prepopulate__ = [
        dict(name='temperature', unit='deg. Celsius'),
        dict(name='light', unit='intensity', comment='Ambient light intensity (can be converted to lux).'),
        dict(name='humidity', unit='%', comment='Relative humidity of the atmosphere.'),
        dict(name='feels_like', unit='deg. Celsius', comment='Feels like temperature is an estimate of the temperature, corrected for humidity.'),
        dict(name='dew_point', unit='deg. Celsius', comment='Dew point temperature.'),
        dict(name='pressure', unit='mbar', comment='air pressure.'),
        dict(name='wind_speed', unit='m/s', comment='Wind speed at 2 meter in m/s.'),
        dict(name='wind_deg', unit='deg.', comment='Wind direct in degree.'),
        dict(name='clouds', unit='%', comment='Cloud coverage in %.')
    ]

    id = Column(Integer, primary_key=True)
    name = Column(String(64), nullable=False, unique=True)
    unit = Column(String(30), nullable=False)
    comment = Column(String)

    raw_data = relationship('RawData', back_populates='variable')
    data = relationship('Data', back_populates='variable')


class Sensor(Base):
    __tablename__ = 'sensors'
    __prepopulate__ = [
        dict(name='OWM observation', comment='OWM observations are a composite product of remote sensing, modelling and in-situ observatrion products.'),
        dict(name='OWM short forecast', comment='OWM forecast < 24 hrs. Composite product of global and local weather models.'),
        dict(name='OWM long forecast', comment='OWM forecast > 24 hrs and < 48 hrs. Composite product of global and local weather models.')
    ]
    
    id = Column(Integer, primary_key=True)
    name = Column(String(), nullable=False)
    comment = Column(String(), nullable=True)

    meta = relationship('Metadata', back_populates='sensor')


class Term(Base):
    __tablename__ = 'terms'

    # columns
    id = Column(Integer, primary_key=True)
    short = Column(String(8), nullable=False)
    full_name = Column(String(), nullable=False)
    start_date = Column(DateTime, nullable=False)
    end_date = Column(DateTime, nullable=False)

    # relationships
    data = relationship('Metadata', back_populates='term')

    @classmethod
    def defaults(cls, session: Session):
        """
        Load some default values
        """
        # build the default values
        cls.__prepopulate__ = []
        for y in range(16, 40):
            y2 = y + 1
            cls.__prepopulate__.extend([
                dict(short='WT%d' % y, full_name='Winterterm 20%d/20%d' % (y, y2), start_date=dt(2000 + y, 10, 1), end_date=dt(2000 + y2, 3, 31)),
                dict(short='ST%d' % y2, full_name='Summerterm 20%d' % y2, start_date=dt(2000 + y2, 4, 1), end_date=dt(2000 + y2, 9, 30)),
            ])
        
        # super into Base class
        super(Term, cls).defaults(session=session)

    def __str__(self):
        return self.short


class Quality(Base):
    __tablename__ = 'quality'
    __prepopulate__ = [
        dict(short='R', name='Regression filled', comment='This measurement was filled by a linear regression model to another station.'),
        dict(short='H', name='Hobo data', comment='This value was quality checked and aggregated from the HOBO data.')
    ]

    id = Column(Integer, primary_key=True)
    short = Column(String(20), nullable=False, unique=True)
    name = Column(String)
    comment = Column(String)


class Data(Base):
    __tablename__ = 'data'

    meta_id = Column(Integer, ForeignKey('metadata.id'), primary_key=True)
    variable_id = Column(Integer, ForeignKey('variables.id'), primary_key=True)
    tstamp = Column(DateTime, primary_key=True)
    value = Column(REAL, nullable=False)
    quality_flag_id = Column(Integer, ForeignKey('quality.id'), nullable=False)

    variable = relationship('Variable', back_populates='data')
    meta = relationship('Metadata', back_populates='data')

class RawData(Base):
    __tablename__ = 'raw_data'
#    __table_args__ = (
#        UniqueConstraint('meta_id', 'variable_id', 'tstamp', name='uq_raw_data')
#    )

    meta_id = Column(Integer, ForeignKey('metadata.id'), primary_key=True)
    variable_id = Column(Integer, ForeignKey('variables.id'), primary_key=True)
    tstamp = Column(DateTime, primary_key=True)
    value = Column(REAL, nullable=False)

    variable = relationship('Variable', back_populates='raw_data')
    meta = relationship('Metadata', back_populates='raw_data')


class Detail(Base):
    __tablename__ = 'details'

    # columns
    id = Column(Integer, primary_key=True)
    key = Column(String, nullable=False)
    str_value = Column(String, nullable=True)
    int_value = Column(Integer, nullable=True)
    float_value = Column(Numeric, nullable=True)
    bool_value = Column(Boolean, nullable=True)

    # relationships
    meta = relationship('Metadata', secondary='nm_metadata_details', back_populates='details')

    @classmethod
    def _field_name(cls, value):
        if isinstance(value, str):
            return 'str_value'
        elif isinstance(value, int):
            return 'int_value'
        elif isinstance(value, float):
            return 'float_value'
        elif isinstance(value, bool):
            return 'bool_value'
        else:
            return 'str_value'
    
    @classmethod
    def get_or_init(cls, session, key, value):
        val_attr = getattr(cls, cls._field_name(value))

        # find
        if session is None:
            instance = None
        else:
            instance = session.query(Detail).filter(Detail.key==key).filter(val_attr==value).first()

        # if Detail does not exist, create it
        if instance is None:
            instance = Detail(key, value)
        
        return instance

    def __init__(self, key, value):
        self.key = key
        setattr(self, self._field_name(value), value)

    def to_dict(self, add_to=None):
        if add_to is None:
            add_to = dict()
        for _t in ['str', 'int', 'float', 'bool']:
            value = getattr(self, '%s_value' % _t)
            if value is not None:
                add_to[self.key] = value
        return add_to


# -----------------------------------------------------------
# Some extra models
# -----------------------------------------------------------
class SpaceRaw(Base):
    __tablename__ = 'space_raw'

    id = Column(BigInteger, primary_key=True)
    company_name = Column(String(65))
    location = Column(String)
    datum = Column(DateTime)
    detail = Column(String)
    status_rocket = Column(String)
    status_mission = Column(String)

class OSMAssociation(Base):
    __tablename__ = 'nm_nodes_tags'

    node_id = Column(BigInteger, ForeignKey('osm_nodes.id'), primary_key=True)
    tag_id = Column(BigInteger, ForeignKey('osm_tags.id'), primary_key=True)

class OSMNode(Base):
    __tablename__ = 'osm_nodes'

    id = Column(BigInteger, primary_key=True)
    geom = Column(Geometry(srid=4326), nullable=False)
    name = Column(String, nullable=True)
    node_type = Column(String, nullable=True)
    raw_tags = relationship('OSMTag', secondary='nm_nodes_tags', back_populates='nodes', cascade='all,delete')

    def get_tags(self):
        return {t.key: t.value for t in self.raw_tags}

    def to_dict(self, stringify=False):
        d = dict(id=self.id, name=self.name, node_type=self.node_type, geom=self.geom)
        tags = self.get_tags()
        
        # stringify
        if stringify:
            d = {k:str(v) for k,v in d.items()}
            tags = {k:str(v) for k,v in tags.items()}

        d['tags'] = tags

        return d


class OSMTag(Base):
    __tablename__ = 'osm_tags'

    id = Column(BigInteger, primary_key=True)
    key = Column(String(2048), nullable=False)
    value = Column(String(2048), nullable=False)
    nodes = relationship('OSMNode', secondary='nm_nodes_tags', back_populates='raw_tags')


class GPX(Base):
    __tablename__ = 'gpx_points'

    id = Column(BigInteger, primary_key=True)
    track_id = Column(Integer, ForeignKey('gpx_tracks.id'), nullable=False)
    geom = Column(Geometry(srid=4326), nullable=False)
    tstamp = Column(DateTime, nullable=False)
    elevation = Column(Numeric, nullable=True)

    track = relationship('GPXTrack', back_populates='points')
    extras = relationship('GPXExtra', back_populates='point')


class GPXTrack(Base):
    __tablename__ = 'gpx_tracks'

    id = Column(Integer, primary_key=True)
    name = Column(String(256), nullable=False)
    tstamp = Column(DateTime, nullable=True)

    points = relationship('GPX', back_populates='track')


class GPXExtra(Base):
    __tablename__ = 'gpx_extra'

    id = Column(BigInteger, primary_key=True)
    gpx_id = Column(BigInteger, ForeignKey('gpx_points.id'), nullable=False)
    tag_name = Column(String(20), nullable=False)
    value = Column(Numeric, nullable=False)

    point = relationship('GPX', back_populates='extras')


class WorldBorder(Base):
    __tablename__ = 'world_borders'

    # columns
    id = Column(Integer, primary_key=True)
    geom = Column(Geometry(srid=4326), nullable=False)
    name = Column(String(1024), nullable=False)
    iso_a2 = Column(String(2), nullable=False)
    iso_a3 = Column(String(3), nullable=False)


import os
import requests
import uuid

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from bs4 import BeautifulSoup

from hydenv.util import env
from hydenv import models

class HydenvGPXExample:
    """
    GPX Example Loader.\n
    Loads GPX example data from Github and load it into the
    given database. The example data url can be overwritten to any 
    remote or local .gpx file. Note, that the loader will only import 
    *flat* GPX data, meaning only the first track will be imported.
    The database has to be installed and initialized before. 
    Do that by hand or use the two CLI commands:
        python -m hydenv database install  --connection=postgresql://postgres:<adminpassword>@localhost:5432/postgres
        python -m hydenv database init --clean --connection=postgresql://hydenv:hydenv@localhost:5432/hydenv
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>
    :param overwrite_url: string - overwrite the default data url to load other data. 
        Can be a remote or local path.
    :param quiet: bool - suppresses print output to stdout
    
    """

    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}", overwrite_url=None, quiet=True):
        # substitute and save
        self.__connection = env.build_connection(connection=connection)
        self.engine = create_engine(self.__connection)
        Session = sessionmaker(bind=self.engine)
        self.session = Session()

        # set some URLs
        if overwrite_url is None:
            self._data_url = 'https://raw.githubusercontent.com/data-hydenv/data/master/extra/spatial/activity/'
        else:
            self._data_url = overwrite_url

        self.quiet = quiet

    def run(self, fname='activity_5967183464.gpx'):
        """
        """
        if os.path.exists(self._data_url + fname):
            content = self.__from_file(fname)
        else:
            content = self.__from_download(fname)

        if not self.quiet:
            print('Parsing GPX...', end='')

        # parsing data
        gpxSoup = BeautifulSoup(content, 'lxml')
        points = gpxSoup.find_all('trkpt')

        # build the data
        try:
            name = gpxSoup.find('name').decode_contents()
        except:
            name = uuid.uuid4()
        tstamp = gpxSoup.metadata.time.decode_contents()
        track = models.GPXTrack(name=name, tstamp=tstamp)

        # get the points
        for point in points:
            track.points.append(self.__parse_point(point))
        
        if not self.quiet:
            print('done.\nUploading...', end='')
        
        try:
            self.session.add(track)
            self.session.commit()
        except Exception as e:
            self.session.rollback()
            if not self.quiet:
                print('error.\n[ERROR]: %s' % str(e))
        
        if not self.quiet:
            print('done.')

    def __from_file(self, fname):
        with open(self._data_url + fname, 'r') as f:
            return f.read()

    def __from_download(self, fname):
        if not self.quiet:
            print('Downloading %s...' % fname, end='')

        # download
        response = requests.get(self._data_url + fname)

        if not response.status_code == 200:
            msg = 'Download failed [%d]\n' % response.status_code
            if not self.quiet:
                print('error.\n[ERROR]: %s' % msg)
            else:
                raise requests.ConnectionError(msg)
        else:
            if not self.quiet:
                print('done.')
        
        return response.content

    def __parse_point(self, p):
        # build data
        wkt = 'srid=4326;POINT ({lon} {lat})'.format(**p.attrs)
        el = float(p.ele.decode_contents())
        tstamp = p.time.decode_contents()

        # create model
        gpx = models.GPX(geom=wkt, tstamp=tstamp, elevation=el)

        # go for extras
        for ex in p.find('ns3:trackpointextension').children:
            if ex.name is None or not ex.name.startswith('ns3:'):
                continue
            gpx.extras.append(models.GPXExtra(
                tag_name=ex.name[4:],
                value=float(ex.decode_contents())
            ))
        
        return gpx



from functools import reduce
import os
import json
from datetime import datetime as dt
from datetime import timedelta as td
from typing import List, Union
import requests
import progressbar
import pandas as pd
from dateutil.parser import parse as date_parse


class HydenvNetatmoExample:
    def __init__(self, password: str = None, quiet: bool = False):
        # set some urls
        self._token_url = "https://us-central1-data-course-261609.cloudfunctions.net/get_token"
        self._netatmo_stations_url = 'https://api.netatmo.com/api/getpublicdata'
        self._netatmo_data_url = 'https://api.netatmo.com/api/getmeasure'

        # set other args
        self._password = password
        self.quiet = quiet
        
        # check for tokens
        self._check_token()

    def _check_token(self):
        # check if there is a netatmo token in environment
        env_token = os.environ.get("NETATMO_TOKEN", "{}")

        token = json.loads(env_token)
        # default to an expired token 
        expires_at = token.get("expires_at", dt.now() - td(days=1))
        
        # check token refresh
        if expires_at > dt.now():
            self._token = token
        else:
            self._token = self._get_token()

    def _get_token(self):
        if self._password is None:
            print("[ERROR] Can't refresh the Bearer token without a password")
            return {}
        
        if not self.quiet:
            print("Requesting a new token...", end="")
        
        # get the new token
        try:
            resp = requests.post(self._token_url, json={'password': self._password}, headers={'Content-Type': 'application/json'})
            resp.raise_for_status()
            print('done.')
        except Exception as e:
            if not self.quiet:
                print("\[ERROR]", str(e))
            raise e
        
        # set the new token to environment
        token = resp.json()
        os.environ['NETATMO_TOKEN'] = json.dumps(token)

        return resp.json()
    
    @property
    def token(self):
        if self._token is None or 'access_token' not in self._token:
            raise AttributeError('No valid Token and no password given to request a new token.')
        else:
            return self._token

    def get_stations(self, lat_ne: float = 48.037, lon_ne: float = 7.89841, lat_sw: float = 47.9569, lon_sw: float = 7.72333, required_data: list = ['temperature'], **kwargs):
        """
        """
        # create the header
        HEADER = {'Authorization': 'Bearer ' + self.token['access_token']}

        # create the params
        PARAMS = dict(lat_ne=lat_ne, lon_ne=lon_ne, lat_sw=lat_sw, lon_sw=lon_sw, required_data=','.join(required_data))

        if not self.quiet:
            print(f"Requesting data at [{self._netatmo_stations_url}] ...", end="")
        
        # do the request
        try:
            response = requests.get(self._netatmo_stations_url, headers=HEADER, params=PARAMS)
            response.raise_for_status()
        except Exception as e:
            if not self.quiet:
                print("faild.\n[ERROR]: ", str(e))
            raise e
        
        if not self.quiet:
            print("done.")
        
        # extract stations
        stations = response.json()['body']
        return stations

    def transform_stations_list(self, stations: list, filter_for=['temperature', 'humidity', 'rain_60min', 'pressure'], **kwargs):
        """
        """
        transformed = []

        for s in stations:
            # get MAC address
            trans = dict(device_id=s['_id'])

            loc = s.get('place', {}).get('location')
            if loc is None:
                print(f"MAC: {trans['device_id']} has no location. Skipping...")
                continue
            else:
                trans['location'] = f"SRID=4326;POINT({loc[0]} {loc[1]})"
            
            # get measures
            m = s.get('measures', {})
            for m_id, measure in m.items():
                for filt in  filter_for:
                    if filt in measure.get('type', []):
                        trans['filt'] = m_id
            
            # append
            transformed.append(trans)

        return transformed

    def load_station_data(self, station: dict, **kwargs) -> dict:
        """
        """
        # get the MAC address
        device_id = station['_id']

        # extract the measures
        measures = station.get('measures', {})
        if not self.quiet:
            print(f"Found {len(measures)} measures on MAC {device_id}.")
        
        # create the header
        HEADER = {'Authorization': 'Bearer ' + self.token['access_token']}

        # get some main settings
        scale = kwargs.get('scale', '30min')

        # check if a start date was set
        start_date = kwargs.get('start', None)
        if start_date is not None:
            start_date = date_parse(str(start_date))
        end_date = kwargs.get('end', None)
        if end_date is not None:
            end_data = date_parse(str(end_date))
        
        # go for each module id
        for module_id, measure in measures.items():
            if 'type' not in measure:
                # TODO I think this is the rainfall module...
                print(f"Found malformed measure: {module_id}. Skipping...")
                continue
            type_ = ','.join(measure['type'])   
            
            # build the params for this round
            PARAMS = dict(device_id=device_id, module_id=module_id, scale=scale, type=type_)

            # check if we use a start data
            if start_date is not None:
                PARAMS['date_begin'] = start_date.timestamp()
            if end_date is not None:
                PARAMS['date_end'] = end_date.timestamp()

            # do the request
            try:
                response = requests.get(self._netatmo_data_url, headers=HEADER, params=PARAMS)
                response.raise_for_status()
                
                # check if data was actually returned
                body = response.json()['body']
                if len(body) == 0:
                    print(f"No body for {device_id}/{module_id}")
                    continue
                elif len(body) > 1:
                    print(f'Weird! Body of {device_id}/{module_id} too large!')
                if len(body[0].get('value')) == 0:
                    print(f"No data found for {device_id}/{module_id}")
                    continue

                # add data
                if 'data' not in station:
                    station['data'] = {}

                # add the data
                station['data'][module_id] = response.json()['body']

            except Exception as e:
                if not self.quiet:
                    print(f"[Error]: Module {module_id} failed: {str(e)}")
                continue
        
        if not self.quiet:
            print(f"Finished loading data for MAC {device_id}.")
        
        return station

    def get_data(self, lat_ne: float, lon_ne: float, lat_sw: float, lon_sw: float, required_data: list = ['temperature'], **kwargs):
        """
        """
        # get the stations
        metadata = self.get_stations(lat_ne=lat_ne, lon_ne=lon_ne, lat_sw=lat_sw, lon_sw=lon_sw, required_data=required_data, **kwargs)

        if not self.quiet:
            print(f"Found {len(metadata)} stations.")
            bar = progressbar.ProgressBar(max_value=len(metadata), redirect_stdout=True)

        stations = []

        # load the data for each stations
        for i, meta in enumerate(metadata):
            station = self.load_station_data(meta, **kwargs)
            stations.append(station)

            if not self.quiet:
                bar.update(i + 1)
        
        if not self.quiet:
            print("Done.")

        # check if stations without data should be removed
        if kwargs.get('clean', False):
            return [station for station in stations if 'data' in station]
        else:
            return stations

    def run(self, bbox=None, city=None, save=None, **kwargs):
        """
        BBOX shape is [lat_ne, lon_ne, lat_sw, lon_sw]
        """
        if bbox is None and city is None:
            print("[ERROR] You need to specify either a bbox or a city.")
            return
        
        CITIES = {
            'fr': [48.037, 7.89841, 47.9569, 7.72333],
            'freiburg': [48.037, 7.89841, 47.9569, 7.72333],
            'ka': [49.06023452689163, 8.541590419576641,48.97413213170277, 8.294625101041413],
            'karlsruhe':  [49.06023452689163, 8.541590419576641,48.97413213170277, 8.294625101041413]
        }
        
        # check for predefined city
        if city is not None:
            if city.lower() not in CITIES.keys():
                print(f"[ERROR] City {city} not pre-defined. Please pass the bbox")
                return
            bbox = CITIES[city.lower()]

        # check for pre-defined bbox
        if bbox is not None:
            lat_ne, lon_ne, lat_sw, lon_sw = bbox
        
        # run
        if kwargs.get('metadata_only', False):
            result = self.get_stations(lat_ne=lat_ne, lon_ne=lon_ne, lat_sw=lat_sw, lon_sw=lon_sw, **kwargs)
        else:
            result = self.get_data(lat_ne=lat_ne, lon_ne=lon_ne, lat_sw=lat_sw, lon_sw=lon_sw, **kwargs)
        
        # check if we want to save to file
        if save is not None:
            with open(save, 'w') as f:
                json.dump(result, f, indent=4)
            if not self.quiet:
                print(f"Saved to {save}.")
        else:
            return result

    def parse(self, fname: Union[str, List[dict]], use_type: str, fmt='dataframe'):
        """
        Pare a saved Netatmo API dump into various formats.\n
        Please note that the parse function can only extract on measure type at
        a time. If you want to extract multiple measure types, you need to
        call the parse function multiple times.
        :param fname: The file name or list of dictionaries to parse.
        :param use_type: The type of measure to extract.
        :param fmt: Output format; ONe of ['dataframe', 'latex', 'markdown', 'csv', 'html']
        """
        # open the file if necessary
        if isinstance(fname, str):
            with open(fname, 'r') as f:
                stations = json.load(f)
        else:
            stations = fname

        data = pd.DataFrame()

        # go for each stations
        for station in stations:
            # get the measures
            measures = station.get('measures', {})
            
            # add search for use type
            for module_id, measure in measures.items():
                if use_type in measure.get('type', []):
                    # get the index of the right measure
                    data_index = measure['type'].index(use_type)
                    raw_data_obj = station.get('data', {}).get(module_id, [[]])[0]
                    values = [obs[data_index] for obs in raw_data_obj.get('value', [])]

                    # build the datetime index
                    index = pd.date_range(start=dt.fromtimestamp(int(raw_data_obj.get('beg_time'))), periods=len(values), freq='30min')
                    df = pd.DataFrame(values, index=index, columns=[f"{station['_id']}__{module_id}"])

                    # merge
                    data = pd.merge(data, df, how='outer', left_index=True, right_index=True)

        # return 
        if fmt == 'dataframe':
            return data
        
        elif fmt == 'markdown':
            return data.to_markdown()
        elif fmt == 'latex':
            return data.to_latex()
        elif fmt == 'csv':
            return data.to_csv()
        elif fmt == 'html':
            return data.to_html()



from random import random, choices, choice
from io import StringIO

import pandas as pd
from sqlalchemy import create_engine
# from barnum import gen_data

from hydenv.util import env

NORMALIZE_SQL ="""
-- PRODUCTS
drop table if exists products cascade;
create table products as
select distinct 
	split_part(split_part(purchase, ' | ', 1), ' - ', 1) as product_id,
	split_part(split_part(purchase, ' | ', 1), ' - ', 2) as name,
    split_part(purchase, ' | ', 1) as orig
from purchases;
alter table products add constraint  pkey_products primary key (product_id);
-- CITIES
drop table if exists cities cascade;
create table cities as
select 
	row_number() over (order by zip) as city_id,
	t.*
from
(select distinct
	split_part(split_part(purchase, ' | ', 2), ' - ', 1) as zip,
	split_part(split_part(purchase, ' | ', 2), ' - ', 2) as city,
	split_part(split_part(purchase, ' | ', 2), ' - ', 3) as state,
    split_part(purchase, ' | ', 2) as orig
from purchases) t;
alter table cities add constraint pkey_cities primary key (city_id);
-- STORES
drop table if exists stores cascade;
create table stores as 
select 
	row_number() over (order by name) as store_id,
	t.name
from (select distinct 
	split_part(purchase, ' | ', 3) as name
from purchases) t;
alter table stores add constraint pkey_stores primary key (store_id);
-- CUSTOMERS
drop table if exists customers;
create table customers as 
select row_number() over (order by customer_id) as id, 
* from (select 
	customer_id,
	prod.product_id,
	s.store_id,
	c.city_id
from purchases p
join products prod on prod.orig=split_part(p.purchase, ' | ', 1)
join stores s on s.name=split_part(p.purchase, ' | ', 3)
join cities c on c.orig=split_part(p.purchase, ' | ', 2)) t;
alter table customers add constraint pkey_customers primary key (id);
alter table customers add constraint fkey_products foreign key (product_id) references products (product_id);
alter table customers add constraint fkey_stores foreign key (store_id) references stores (store_id);
alter table customers add constraint fkey_cities foreign key (city_id) references cities (city_id);
COMMIT;
"""
DROP_SQL = """
drop table if exists products cascade;
drop table if exists cities cascade;
drop table if exists stores cascade;
drop table if exists customers;
"""
class HydenvCustomerExamples:
    """
    Space Missions Example Loader.\n
    Loads Space Missions example data from Github and load it into the
    given database. The database has to be installed and initialized before. 
    Do that by hand or use the two CLI commands:
        python -m hydenv database install  --connection=postgresql://postgres:<adminpassword>@localhost:5432/postgres
        python -m hydenv database init --clean --connection=postgresql://hydenv:hydenv@localhost:5432/hydenv
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>

    """
    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}"):
        # substitute and save
        self.__connection = env.build_connection(connection=connection)
        self.engine = create_engine(self.__connection)

    def run(self, quiet=True, k=500, normalize=False):
        """
        """
        pass
        # 3% different stores - max 60
        # maxs = 60 if int(0.03 * k) > 60 else int(0.03 * k) 
        # stores = [gen_data.create_company_name() for _ in range(maxs)]

        # # 2% different locations - max 21
        # maxs = 21 if int(0.02 * k) > 21 else int(0.02 * k)
        # locations = [gen_data.create_city_state_zip() for _ in range(maxs)]

        # # 8% different products - 1200 max 
        # maxs = 1200 if int(0.08 * k) > 1200 else int(0.08 * k)
        # products = [(int(random() * 10**8), gen_data.create_nouns(choices([1,2,3], weights=[0.7, 0.2, 0.1], k=1)[0])) for _ in range(maxs)]

        # # 60% different customers
        # customer_ids = choices(list(range(int(0.6 * k))), k=k)

        # csv = "customer_id,purchase\n"

        # for customer_id in customer_ids:
        #     # new dataset
        #     p = '%s - %s' % choice(products)
        #     loc = ' - '.join(choice(locations))
        #     line = '%s,%s' % (customer_id, ' | '.join([p, loc, choice(stores)]))
            
        #     csv += line + '\n'
        
        # # build buffer
        # buf = StringIO()
        # buf.write(csv)
        # buf.seek(0)

        # df = pd.read_csv(buf)

        # if not quiet:
        #     print('Uploading...')
        
        # # save to database
        # with self.engine.connect() as con:
        #     con.execute('DROP TABLE IF EXISTS purchases CASCADE;COMMIT;')
        # df.to_sql('purchases', self.engine, index=None, if_exists='replace')

        # if normalize:
        #     if not quiet:
        #         print('Normalizing data scheme...')
        #     self._normalize()

        # if not quiet:
        #     print('Done.')

    def _normalize(self):
        with self.engine.connect() as con:
            con.execute(NORMALIZE_SQL)

    def drop(self):
        with self.engine.connect() as con:
            con.execute(DROP_SQL)



from typing import List
import requests
import zipfile
import io
import os
import shutil
import glob
import json
import hashlib
from collections import defaultdict
import pandas as pd
from datetime import datetime as dt

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from hydenv.util import env
from hydenv import models


class HydenvOWMExample:
    """
    OWM Example Loader.\n
    Loads OWM example data from Github (or other sources) and downloads them as
    CSV files. These CSV files can then be loaded into the given database. 
    The database has to be installed and initialized before. 
    Do that by hand or use the two CLI commands:
        python -m hydenv database install  --connection=postgresql://postgres:<adminpassword>@localhost:5432/postgres
        python -m hydenv database init --clean --connection=postgresql://hydenv:hydenv@localhost:5432/hydenv
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>
    """
    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}"):
        # substitute the connection string
        self.__connection = env.build_connection(connection=connection)
        self.engine = create_engine(self.__connection)

        # set some urls
        self._data_url = "https://github.com/data-hydenv/data/archive/master.zip"
        
    def _download_data(self):
        # download the zip into memory
        r = requests.get(self._data_url)
        z = zipfile.ZipFile(io.BytesIO(r.content))

        # save to current directory
        z.extractall()

    def _read_all(self, path: str, pattern: str):
        # empty container
        meta = {}
        observ = {}
        forecast = defaultdict(lambda: list())

        # read all files in the given path
        files = glob.glob(os.path.join(os.path.join(path, pattern)))

        for f in files:
            # read the file
            with open(f, 'r') as f:
                data = json.load(f)
            
            # get the filename
            fname = os.path.basename(f.name).split('_raw_dump')[0]
            tstamp = dt.strptime(fname, '%Y_%m_%d')

            # add observations
            obs = data.get('historic', {}).get('hourly', [])
            for o in obs:
                observ[o['dt']] = o

            # add forecasts
            pred = data.get('forecast', {}).get('hourly', [])
            for p in pred:
                p.update({'from_dt': int(tstamp.timestamp())})
                forecast[p['dt']].append(p)
            
            # update the meta
            meta.update({k: v for k, v in data.get('historic', {}).items() if k not in ('hourly', 'current')})
        
        # return
        return meta, observ, forecast

    def _order_to_ts(self, observations: dict, forecast: dict) -> List[dict]:
        # create the master ts 
        index = list(observations.keys())
        #index.extend([ts for ts in forecast.keys() if ts not in index])

        # sort
        index.sort()
        
        # create the data container
        data = []

        for idx in index:
            obs = observations[idx]
            obs['forecast'] = forecast.get(idx, [])
            data.append(obs)
        
        return data
    
    def _upload_to_database(self, df: pd.DataFrame, key='temp', meta: dict = {}, if_exists: str = 'replace') -> None:
        """
        """
        # temp is called 'temperature' in the database
        if key == 'temp':
            var_key = 'temperature'
        else:
            var_key = key
        
        # create a database session
        Session = sessionmaker(bind=self.engine)
        session = Session()

        # get the variable
        variable = session.query(models.Variable).filter(models.Variable.name==var_key).one()

        # get the location
        lon = meta.get('lon')
        lat = meta.get('lat')

        KEYS = {'OWM observation': key, 'OWM short forecast': 'forecast_1_value', 'OWM long forecast': 'forecast_2_value'}
        df.reset_index(inplace=True)

        # go for the three products
        for prod in ['OWM observation', 'OWM short forecast', 'OWM long forecast']:
            # get the sensors
            sensor = session.query(models.Sensor).filter(models.Sensor.name == prod).one()

            # derive the device IDs
            md5 = hashlib.md5(f'{prod}{lon} {lat}'.encode()).hexdigest()
            device_id = f'OWM_{md5}'

            # check if the metadata exists
            metadata = session.query(models.Metadata).filter(models.Metadata.device_id==device_id).first()
            if metadata is None:
                metadata = models.Metadata(
                    device_id=device_id,
                    sensor_id=sensor.id,
                    location=f'SRID=4326;POINT ({lon} {lat})' if lon is not None and lat is not None else None,
                )
                metadata.set_details(details=meta, session=session)

                # save the metadata
                session.add(metadata)
                session.commit()
            elif if_exists == 'raise':
                raise ValueError('The metadata already exists')
            elif if_exists == 'replace':
                with session.bind.connect() as con:
                    con.execute(f'DELETE FROM raw_data WHERE meta_id={metadata.id};COMMIT;')

            # add the data
            data = df[['time', KEYS[prod]]].copy()
            data.columns = ['tstamp', 'value']
            
            data['meta_id'] = metadata.id
            data['variable_id'] = variable.id
            
            # drop NA values
            data.dropna(how='any', inplace=True)

            # check for duplicates
            imp_data = data.drop_duplicates(['tstamp', 'meta_id', 'variable_id']).copy()
            if len(imp_data) != len(data):
                print(f'\n[WARNING]: duplicated rows for {prod} of KEY {key}')
                data = imp_data
            
            # upload
            data.to_sql('raw_data', session.bind, index=None, if_exists='append')
        
        # close the database session again
        session.close()
    
    def _ordered_data_to_df(self, data: List[dict], key='temp') -> pd.DataFrame:
        # extract only the data we need
        # temp = [{'time': dt.fromtimestamp(d['dt']), key: d[key]} for d in data]
        temp = []

        for d in data:
            # get the main index
            obj = {'time': dt.fromtimestamp(d['dt']), key: d[key]}

            # add the forecasts
            deltas = []
            values = []
            for fore in d['forecast']:
                delta_hrs = int((obj['time'] - dt.fromtimestamp(fore['from_dt'])).total_seconds() / 3600)
                deltas.append(delta_hrs)
                values.append(fore[key])

            # sort the deltas and values
            values = [v for _, v in sorted(zip(deltas, values))][:2]
            deltas = sorted(deltas)[:2]
            
            # add to the main obj
            obj['forecast_1_value'] = values[0] if len(values) >= 1 else None
            obj['forecast_1_delta [hrs]'] = deltas[0] if len(deltas) >= 1 else None
            obj['forecast_2_value'] = values[1] if len(values) >= 2 else None
            obj['forecast_2_delta [hrs]'] = deltas[1] if len(deltas) >= 2 else None
            
            # append
            temp.append(obj)

        # convert
        df = pd.DataFrame(temp)
        df.set_index('time', inplace=True)

        return df

    def run(self, save: str = None, fmt='json', variable=['temp', 'humidity'], if_exists='replace', repo_path='extra/weather/data', pattern='*_raw_dump.json', dry=False, quiet=True):
        # handle variables
        VARIABLES = ['temp', 'feels_like', 'pressure', 'humidity', 'dew_point', 'wind_speed', 'wind_deg', 'clouds']
        if variable == 'all':
            variable = VARIABLES
       
        # check if all variables are allowed
        if isinstance(variable, str):
            variable = [variable]
        for v in variable:            
            if v not in VARIABLES:
                raise ValueError('The variable {} is not allowed'.format(v))
        
        # first download the data
        if not quiet:
            print("Downloading data...", end="")
        self._download_data()
        if not quiet:
            print("done.")
        
        # get the path inside the repo
        data_path = os.path.join('./data-master', repo_path)

        # read all files
        if not quiet:
            print('Converting data...', end='')
        meta, observ, forecast =  self._read_all(path=data_path, pattern=pattern)
        data = self._order_to_ts(observations=observ, forecast=forecast)
        if not quiet:
            print('done.')

        # finally remove all
        if os.path.exists('./data-master'):
            shutil.rmtree('./data-master')

        # upload if needed
        if save is None and dry is False:
            for v in variable:
                if not quiet:
                    print(f'Uploading {v}...', end='')
                
                # handle upload
                df = self._ordered_data_to_df(data=data, key=v)
                self._upload_to_database(df=df, key=v, meta=meta, if_exists=if_exists)
            
                if not quiet:
                    print('done.')
            return

        # save or return it
        if fmt == 'json':
            if dry:
                return data
        
            if not save.endswith('.json'):
                save += '.json'
                with open(save, 'w') as f:
                    json.dump({
                        'metadata': meta,
                        'data': data
                    }, f, indent=4)
            if not quiet:
                print("Saved data to {}".format(save))

        
        elif fmt == 'csv':
            # check how to return
            if dry:
                dfs = {v: self._ordered_data_to_df(data, key=v) for v in variable}
                if len(variable) == 1:
                    return dfs[variable[0]]
                else:
                    return dfs
                
            # save
            for v in variable:
                _save = save
                if _save.endswith('.csv'):
                    _save = _save.replace('.csv', '_' + v + '.csv')
                else:
                    _save = f'{_save}_{v}.csv'
            
                df = self._ordered_data_to_df(data, key=v)
                df.to_csv(_save)
                if not quiet:
                    print("Saved data to {}".format(_save))



import json

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from overpy import Overpass
from overpy.exception import DataIncomplete, OverpassTooManyRequests, OverpassBadRequest
from progressbar import ProgressBar
from shapely.geometry import LineString, Polygon, MultiPolygon
from shapely.ops import linemerge, polygonize, unary_union

from hydenv.util import env
from hydenv import models

# set some helpful overpass API queries
BY_AREA = """
area[boundary~"administrative"][name="{boundary_name}"];
{obj}(area){filter};(._;>;);out;
"""

# some lookups
FEDERAL_STATES = [
    'Baden-Württemberg', 'Bayern', 'Rheinland-Pflaz', 'Saarland', 'Hessen',
    'Nordrhein-Westfalen', 'Niedersachsen', 'Bremen', 'Hamburg', 'Schleswig-Holstein',
    'Thüringen', 'Sachsen', 'Sachsen-Anhalt', 'Brandenburg', 'Berlin', 'Mecklenburg-Vorpommern'
]

CITIES = dict(KA='Karlsruhe', FR='Freiburg im Breisgau', S='Stuttgart',
              B='Berlin', M='München', MA='Mannheim')


class HydenvOSMExamples:
    """
    OSM Example Loader.\n
    Loads Open Street Map example either as persisted dumps from Github, or 
    dynamically direclty from the OSM database using Overpass API. The data 
    is then loaded into the given database. The database has to be installed 
    and initialized before. 
    Do that by hand or use the two CLI commands:
        python -m hydenv database install  --connection=postgresql://postgres:<adminpassword>@localhost:5432/postgres
        python -m hydenv database init --clean --connection=postgresql://hydenv:hydenv@localhost:5432/hydenv
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>

    """
    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}", overpass_url=None):
        # substitute and save
        self.__connection = env.build_connection(connection=connection)
        self.engine = create_engine(self.__connection)
        Session = sessionmaker(bind=self.engine)
        self.session = Session()

        # set the Overpass API
        self.api = Overpass(url=overpass_url)

    def _build_query(self, geometry, boundary, attribute, value=None):
        # build the filter
        if isinstance(attribute, str):
            filt = '[\"%s\"%s]' % (attribute, '~%s' % value if value is not None else '')
        elif isinstance(attribute, dict):
            filt = ''.join(['["%s"%s]' % (attr, '~%s' % val if val is not None else '') for attr, val in attribute.items()])
        
        # build the query
        query = BY_AREA.format(obj=geometry, boundary_name=boundary, filter=filt)
        return query

    def run(self, geometry, boundary, attribute, value=None, dry=False, save=None, type_alias=None, quiet=True):
        # first build the query
#        value = '~%s' % value if value is not None else ''
#        query = BY_AREA.format(obj=geometry, boundary_name=boundary, attribute=attribute, value=value)
        
        # check if boudary was pre-defined
        if boundary in CITIES.keys():
            boundary = CITIES[boundary]
        
        query = self._build_query(geometry, boundary, attribute, value=value)

        # if dry run, do not reach out to the Overpass server
        if dry:
            if not quiet:
                return query
            else:
                print('Overpass QL:\n%s' % query)
                return
        
        # get the data
        if not quiet:
            print('Connecting %s...' % self.api.url)
        try:
            result = self.api.query(query)

            # handle nodes
            if geometry == 'nodes' or geometry == 'node':
                nodes = result.get_nodes()
            
            # handle relations
            elif geometry == 'relation':
                nodes = []
                for relation in result.relations:
                    # get all outer ways
                    mems = [m.resolve().get_nodes() for m in relation.members if m.role == 'outer']
                    nodes.append((relation, mems))

            # handle anything that can be collapsed to ways
            else:
                nodes = result.ways
        except OverpassBadRequest: 
            print("[Error]: Your query is wrong!\nOverpass QL:\n%s" % query)
            return
        except Exception as e:
            print('The following error was caused during Overpass connection:\n%s' % str(e))
            return

        # inform
        if not quiet:
            print('Overpass QL:\n%s\nReturned %s raw objects' % (query, len(nodes)))


        # inform and create the progressbar
        if not quiet:
            print('Uploading...')
            bar = ProgressBar(max_value=len(nodes), redirect_stdout=True)
        
        if save is not None:
            json_objects = []
        for i, node in enumerate(nodes):
            # Point nodes
            if geometry == 'nodes' or geometry == 'node':
                wkt = 'SRID=4326;POINT (%f %f)' % (float(node.lon), float(node.lat))

            # Relations
            elif geometry == 'relation':
                ways = node[1]    # Nodes in the relation
                merged = linemerge([LineString([(p.lon, p.lat) for p in w]) for w in ways])
                parts = list(polygonize(unary_union(merged)))
                poly = parts[0] if len(parts) == 1 else MultiPolygon(parts)
                try:
                    # with shapely > 1.8 this does not work anymore
                    wkt = 'SRID=4326;%s' % poly.to_wkt() 
                except AttributeError:
                    wkt = f"SRID=4326;{poly.wkt}"

                node = node[0]      # the relation itself
            
            # Way-like
            else:
                try:
                    points = node.get_nodes(resolve_missing=False)
                except DataIncomplete:
                    points = node.get_nodes(resolve_missing=True)
                except OverpassTooManyRequests:
                    print('Too many requests to Overpass API, maybe you need to use a smaller boundary.')
                    continue
                # check if a POLYGON or LINESTRING is needed
                if points[0] == points[-1]:
                    wkt = 'SRID=4326;POLYGON ((%s))' % ','.join(['%s %s ' % (p.lon, p.lat) for p in points])
                else:
                    wkt = 'SRID=4326;LINESTRING (%s)' % ','.join(['%s %s' % (p.lon, p.lat) for p in points])

            # check if there is a type alias
            if type_alias is None:
                type_name = value if value is not None and value != '' else attribute
            else:
                type_name = type_alias
            
            # upload
            osm = models.OSMNode(
                id=node.id, 
                geom=wkt, 
                name=node.tags.get('name'),
                node_type=type_name
            )
            for k,v in node.tags.items():
                osm.raw_tags.append(models.OSMTag(key=k, value=v))

            # upload
            if save is None:
                try:
                    self.session.add(osm)
                    self.session.commit()
                except Exception as e:
                    self.session.rollback()
                    if not quiet:
                        print('[%d]: %s' % (osm.id, str(e)))
                    continue
            else:
                json_objects = osm.to_dict(stringify=True)
                    
            if not quiet:
                bar.update(i + 1)
        
        # save if we are in save mode
        if save is not None:
            with open(save, 'w') as f:
                json.dump(json_objects, f, indent=4)

        if not quiet:
            print('Done')
        

    def energiewende(self, boundary="Baden-Württemberg", quiet=True):
        # check if this is a predefined city
        if boundary in CITIES:
            boundary = CITIES[boundary]
        
        # Get the gas stations 
        if not quiet:
            print('Loading all Gas stations in %s' % boundary)
        self.run('node', boundary, 'amenity', 'fuel', quiet=quiet)

        # get the charging stations
        if not quiet:
            print('Loading all charging stations in %s' % boundary)
        self.run('node', boundary, 'amenity', 'charging_station', quiet=quiet)

        # if a known boundary is chose, add counties or districts
        if boundary in FEDERAL_STATES:
            self.counties(state=boundary, quiet=quiet)
        elif boundary in CITIES:
            self.city_districts(city=boundary, quiet=True)

        if not quiet:
            print('Energiewende loaded.')

    def city_districts(self, city="Ka", save=None, quiet=True):
        boundary = CITIES.get(city, city)
        
        if not quiet and city not in CITIES:
            print('%s is not one of our prepared cities abbreviations. The example might still work but is used on your own responsibility' % city )
        if not quiet:
            print('Loading all administrative boundaries of %s' % boundary)
        
        self.run('relation', boundary, 'admin_level', '"9|10"', save=save, type_alias='district', quiet=quiet)
        
        if not quiet:
            print('Done.')

    def counties(self, state="Baden-Württemberg", save=None, quiet=True):
        if not quiet:
            print('Loading all counties in the state %s' % state)
        self.run('relation', state, attribute=dict(boundary='administrative', admin_level='6'), save=save, type_alias='county', quiet=quiet)



import glob
import os
import sys
import json
import hashlib
import time
from io import StringIO
from datetime import datetime as dt
from datetime import timedelta as td
from random import choice
from string import ascii_letters

import streamlit as st
import extra_streamlit_components as stx
from streamlit_ace import st_ace, THEMES
from universal_analytics import Tracker, HTTPRequest

from hydenv.database import HydenvDatabase
from hydenv.examples.examples import HydenvExamples
from hydenv.examples.osm import CITIES, FEDERAL_STATES

BASEPATH = os.path.dirname(os.path.abspath(__file__))
WELCOME = """
This application offers a GUI to some parts of the hydenv CLI. The whole software was developed by hydrocode.
The software is open source, but is able to connect resources, content and data, that are technically not part
of the software. They can be used free of charge for educational purposes. 
Remote database connection and cloud saving of exercises are excluded from these terms and might need registration and/or 
subscription in the future.

The application tracks some general user behavior using Google analytics. The IP adress is anonymized and can not be tracked
back to you. The data collected is: Your OS version, browser version and solving attempts of exercises. Following Art. 6 par. 1
GDPR we have a legitimate interest to verify technical integrety of the application. The duration of saving is 2 months and 
the location is within Europe. By using this application, you consent to these terms.
"""

EXAMPLE_INTRO ="""
The hydenv CLI can automatically download, merge, prepare and filter remote datasets for you.
It also includes database models and can upload the data samples into the connected database.
There are many different dataset that can be loaded. Possibly, not all of the examples can 
yet be run in the GUI.
"""

CONNECT_INTRO = """To use the hydenv GUI, you need to connect a PostgreSQL database.
The easiest way is to install it locally and create a new instance here, using a superuser.
That is a one-time setup to create a new user and all needed tables, populated with some example data.

If the instance is already initialized, you can choose the 'connect existing db' option and set the connection information
of the newly created user.

It is also possible to connect one of the remote databases, which are hosted by hydrocode. In these cases, you will have received
the connection information already.
"""

def get_base_data():
    # get the tracks
    with open(os.path.join(BASEPATH, "maintain/tracks.json")) as f:
        tracks_ = json.load(f)
    
    # get the exercises
    exercises_ = list()
    files = glob.glob(os.path.join(BASEPATH, "maintain", "exercises*.json"))
    for file_ in files:
        with open(file_, 'r') as f:
            exercises_.extend(json.load(f))
    
    tracks = []
    # merge the lists
    for track in tracks_:
        sessions = []
        for session_ in track.get('sessions', []):
            exercises = []
            for exercise in session_.get('exercises', []):
                # find the exercise
                ex_data = [e for e in exercises_ if e.get('id') == exercise.get('id')].pop()
                exercise.update({k: v for k, v in ex_data.items() if k!='id'})
                
                # add
                exercises.append(exercise)
            session_.update({'exercises': exercises})
            sessions.append(session_)
        track.update({'sessions': sessions})
        tracks.append(track)    
    
    return tracks


def manage_session(tracks) -> dict:
    # load the track id
    track_id = st.session_state.get('track_id', False)

    # load the session id
    session_id = st.session_state.get('session_id', False)

    if not session_id or not track_id:
        # select a track
        TRACK = {t['id']: t['name'] for t in tracks}
        track_id = st.selectbox("Track", options=list(TRACK.keys()), format_func=lambda k: TRACK.get(k))

        # select a session
        track = [t for t in tracks if t['id'] == track_id].pop()
        SESSIONS = {s['name']: s['name'] for s in track.get('sessions', [])}
        session_id = st.selectbox("Session", options=list(SESSIONS.keys()), format_func=lambda k: SESSIONS.get(k))

        save = st.button("SELECT")
        if save:
            st.session_state.track_id = track_id
            st.session_state.session_id = session_id
            st.experimental_rerun()
        else:
            st.stop()

    # load the correct exercise
    exercise = st.session_state.get('exercise', None)
    if exercise is None:
        session = get_session(tracks, track_id, session_id)
        for e in session.get('exercises', []):
            if e['id'] in st.session_state.get('solved_exercises', []):
                continue
            else:
                exercise = e
                st.session_state.exercise = exercise
    if exercise is None:
        st.stop()
    return exercise


def get_session(tracks, track_id, session_id=None):
    # get the track
    track = [t for t in tracks if t['id'] == track_id].pop()
    if session_id is None:
        return track
    
    # get the session
    session = [s for s in track.get('sessions', []) if s['name'] == session_id].pop()
    return session


def load_solved_list():
    # TODO use firebase for paid accounts
    cookie_manager = stx.CookieManager()
    solved = cookie_manager.get('solved_exercises')
    if solved is None:
        solved = []
    st.session_state.solved_exercises = solved


def reset_track():
    if 'track_id' in st.session_state:
        del st.session_state.track_id
    if 'session_id' in st.session_state:
        del st.session_state.session_id
    if 'exercise' in st.session_state:
        del st.session_state.exercise


def change_exercise(tracks, direction='next'):
    # get the current exercise
    session = get_session(tracks, st.session_state.track_id, st.session_state.session_id)
    exercises = session.get('exercises', [])
    ex_id = st.session_state.exercise['id']
    idx = [i for i, e in enumerate(exercises) if e['id']==ex_id].pop()

    # check if it was the last one
    if direction == 'next':
        if len(exercises) == idx + 1:
            idx = 0
        else:
            idx += 1
    else:
        idx = idx - 1 if idx > 1 else 0
    
    st.session_state.exercise = exercises[idx]
    st.experimental_rerun()


def exercise_page(tracks, db: HydenvDatabase):
    # reset button
    reset = st.sidebar.button("Change Track & Session")
    if reset:
        reset_track()
        st.experimental_rerun()

    # main container
    #main = st.container()
    editor = st.sidebar.expander('EDITOR', expanded=True)
    layout = editor.radio('LAYOUT', options=['split', 'column'])
    if layout == 'split':
        main, right = st.columns(2)
    else:
        main = st.container()
        right = st.container()

    # get the exercise
    exercise = st.session_state['exercise']
    
    # build the page
    main.title(exercise['name'])
    body = main.expander('BODY', expanded=True)
    body.write(exercise['body'].get('content', '<i>No description found</i>'), unsafe_allow_html=True)

    # hint
    if 'hint' in exercise:
        hint = main.expander('HINT', expanded=False)
        hint.write(exercise['hint'], unsafe_allow_html=True)

    right.markdown('## Results')
    # main area
    with main.form('SQL input'):
        # check if there is prefill
        prefill = st.session_state.get(f"{exercise['id']}_prefill", exercise['body'].get('prefill', ''))
        
        code_theme = editor.selectbox('Editor theme', options=THEMES, index=2)
        size = editor.selectbox('Editor Size', options=['sm', 'md', 'lg'])
        sql_code = st_ace(
            value=prefill,
            placeholder='Put your SQL here', 
            height=600 if size == 'lg' else 450 if size == 'md' else 200,
            language='sql', 
            theme=code_theme, 
            auto_update=True
        )
        # compare mode


        do_run = st.form_submit_button('EXECUTE')

        if do_run:
            # save the query
            st.session_state[f"{exercise['id']}_prefill"] = sql_code
            
            # run
            result = db.execute(sql_code, json=True)
            solution = db.execute(exercise['solution']['content'], json=True)
            
            # save
            st.session_state.last_run = dict(exercise_id=exercise['id'], result=result)

            # here check if result correct
            res_hash = hashlib.sha256(str(result).encode()).hexdigest()
            sol_hash = hashlib.sha256(str(solution).encode()).hexdigest()
            main.code(f'Your hash:\t{res_hash}\nSolution hash:\t{sol_hash}')
            
            if res_hash == sol_hash:
                event('event', category='exercise_attempt', label=exercise['id'], exercise_id=exercise['id'], solved=True)
                right.success("**Great!** Your soultion was correct.")
                
                # TODO do more
                solved = st.session_state.get('solved_exercises', [])
                if exercise['id'] not in solved:
                    solved.append(exercise['id'])
                    
                    # update session state
                    st.session_state.solved_exercises = solved
                    cookie_manager = stx.CookieManager()
                    cookie_manager.set('solved_exercises', solved, expires_at=dt.now() + td(days=30), key='set_solved')
                    time.sleep(0.3)
            else:
                event('event', category='exercise_attempt', label=exercise['id'], exercise_id=exercise['id'], solved=False)
                right.error(f"Argh!. Not yet.")
    
    # indicate if this exercise has been solved before
    if exercise['id'] in st.session_state.solved_exercises:
        right.success('This task has already been solved')
    else:
        right.info('This task was not yet solved.')
    
    # right column
    run = st.session_state.get('last_run', {})
    if run.get('exercise_id', '-99') == exercise['id']:
        right.dataframe(run['result'])
        # right.json(run['result'])

    # solution
    solution_exp = main.expander('SOLUTION', expanded=False)
    solution_exp.code(exercise['solution'].get('content', ''), language='sql')
    
    # navigate buttons
    l, _, r = st.columns((1, 3, 1))
    go_prev = l.button('PREVIOUS EXCERCISE')
    go_next = r.button('NEXT EXCERCISE')
    if go_prev or go_next:
        change_exercise(tracks, direction='next' if go_next else 'previous')


def get_db_con(connection: str = None) -> HydenvDatabase:
    # check if the connection has peviously been set
    con = st.session_state.get('connection')
    if con is not None:
        return HydenvDatabase(connection=con)

    else:
        install_page()
    # in any other case
    db = _check_connection(connection=connection)
    if db:
        # if that worked, set the connection
        st.session_state.connection = db.unsafe_get_connection
        st.experimental_rerun()
    else:
        # force the install page
        install_page()


def _check_connection(connection: str = None, with_tables: bool = True):
    try:
        db = HydenvDatabase(connection=connection)
        if with_tables:
            res = db.execute('SELECT * FROM space_raw LIMIT 1;')
        else:
            res = db.execute('SELECT * FROM information_schema.tables LIMIT 1;')
        return db
    except Exception:
        return False


def install_page():
    # cookie consent
    if not st.session_state.get('consent', False):
        st.markdown('### A hydrocode application')
        st.markdown('# Hydenv GUI')
        st.markdown(WELCOME)

        got_it = st.button('Continue')
        if got_it:
            st.session_state.consent = True
            st.experimental_rerun()
        else:
            st.stop()
    
    # there is nothing to try - could use a cookie here
    st.markdown('## Connect hydenv database')
    st.markdown(CONNECT_INTRO)
            
    # check if there was a error message
    if 'error_msg' in st.session_state:
        st.error(st.session_state.error_msg)
        del st.session_state.error_msg
    
    if 'success_msg' in st.session_state:
        st.success(st.session_state.success_msg)
        del st.session_state.success_msg

    if not st.session_state.get('config_connection', False):
        # build the 
        left, right, _ = st.columns((1, 1, 4))
        install_new = left.button('CREATE NEW HYDENV INSTANCE')
        connect_existing = right.button('CONNECT EXISTING HYDENV DB')
        
        if install_new:
            st.session_state.config_connection = 'install_new'
        elif connect_existing:
            st.session_state.config_connection = 'connect_existing'
        else:
            st.stop()

    if st.session_state.config_connection == 'finished':
        del st.session_state.config_connection
        msg = st.empty()
        bar = st.progress(0)

        for i in range(5):
            msg.write("You will be redirected in {} seconds".format(5 - i))
            bar.progress((i + 1) *20)
            time.sleep(1)
        
        st.session_state.connection = st.session_state.raw_connection
        del st.session_state.raw_connection
        st.session_state.page_name = 'home'
        st.experimental_rerun()

    # build the menu
    with st.form('SETUP CONNECTION'):
        host = st.text_input('DATABASE HOST', 'localhost')
        port = st.text_input('PORT', '5432')
        if st.session_state.config_connection == 'install_new':
            pg_pass = st.text_input('POSTGRES SUPER PASSWORD')
        user = st.text_input('HYDENV USER', 'hydenv')
        userpw = st.text_input('HYDENV PASSWORD', 'hydenv')
        db_name = st.text_input('DATABASE NAME', 'hydenv')
        submit = st.form_submit_button()

        # check 
        if submit:
            if st.session_state.config_connection == 'install_new':
                con = f"postgresql://postgres:{pg_pass}@{host}:{port}/postgres"
                
                # do the stuff
                try:
                    db = HydenvDatabase(connection=con)
                    with st.spinner('Creating database...'):
                        db.install_silent(db_name=db_name, user=user, password=userpw, skip_init=False)

                    # finished
                    st.session_state.raw_connection = db.unsafe_get_connection
                    st.session_state.config_connection = 'finished'
                    st.session_state.success_msg = 'Successfully created database. You will be redirected to the main page.'
                    st.experimental_rerun()
                except Exception as e:
                    st.session_state.error_msg = str(e)
                    st.experimental_rerun()
            elif st.session_state.config_connection == 'connect_existing':
                con = f"postgresql://{user}:{userpw}@{host}:{port}/{db_name}"
                
                # check if it works with tables
                if _check_connection(connection=con):
                    st.session_state.raw_connection = con
                    st.session_state.config_connection = 'finished'
                    st.experimental_rerun()
                
                # if not try to install the tables or fail
                else:
                    try:
                        db = _check_connection(connection=con, with_tables=False)
                        if not db:
                            raise Exception('Could not connect to database')
                        db.init(clean=True)
                        st.session_state.raw_connection = db.unsafe_get_connection
                        st.session_state.config_connection = 'finished'
                        st.session_state.success_msg = 'Successfully connected to database. You will be redirected to the main page.'
                        st.experimental_rerun()
                    except Exception as e:
                        st.session_state.error_msg = str(e)
                        st.experimental_rerun()
            else:
                st.error('Something went wrong')
                st.json(st.session_state)
                st.stop()
    # STOP
    st.stop()


def example_page(db: HydenvDatabase):
    example = st.session_state.page_name.split('_')[1]

    # CHECK queries
    CHECK = {
        'hobo': ["SELECT short as \"Identifier\", full_name as \"Term\", count(*) AS \"HOBO Metadata\" FROM metadata m JOIN terms ON terms.id=m.term_id JOIN sensors s ON m.sensor_id=s.id WHERE s.name='hobo' GROUP BY full_name, short;"],
        'osm': [
            """SELECT node_type as "Node Type", count(*) as "Features" from osm_nodes GROUP BY node_type;""",
            """select 'Distinct OSM tags' as "Description", count(distinct key) from osm_tags union select 'Total tags on Nodes' as "Description", count(*) from nm_nodes_tags;"""
        ]
    }
    CMD = {
        'hobo': 'python -m hydenv examples hobo',
        'osm': 'python -m hydenv examples osm',
        'netatmo': 'python -m hydenv examples netatmo'
    }
    cmd = CMD[example]

    # show the title
    st.title('Example {}'.format(example.upper()))
    
    # show the tables
    if example in CHECK:
        st.markdown('### Existing tables\n Please make sure that the examples data is not listed below. Most example APIs will create dublicates if you run them twice.')
        for query in CHECK[example]:
            overview_data = db.execute(query, json=True)
            st.table(overview_data)

    st.sidebar.markdown('### API options')
    
    # example-specific code
    is_defnied = True
    args = {}
    if example == 'hobo':
        # terms
        terms = st.sidebar.multiselect('Terms', options=['WT18', 'WT19', 'WT20', 'WT21', 'WT22'])
        if len(terms) == 0:
            st.warning('It is recommended to explicitly select the terms. Otherwise all terms will be selected.')
        else:
            args = {'terms': terms}
        
        # handle only flag
        ONLYS = {'all':'Import all at once', 'metadata': 'HOBO Metadata', 'raw-data': 'HOBO raw data', 'quality-data': 'HOBO quality checked data'}
        only = st.sidebar.radio('Import only', options=list(ONLYS.keys()), index=0, format_func=lambda k: ONLYS[k])

        if only != 'all':
            args['only'] = only
    
    elif example == 'osm':
        # action
        action = st.sidebar.selectbox('Action', options=['city-districts', 'counties', 'energiewende', 'nodes', 'way'])
        args = {'action': action}

        if action not in ('nodes', 'way'):
            use_predefined = st.sidebar.checkbox('Use predefined geometry settings', value=True)
        
        if action == 'city-districts':
            if use_predefined:
                city = st.sidebar.selectbox('PREDEFINED CITY', options=list(CITIES.keys()), format_func=lambda k: CITIES[k])
                args['city'] = city
            else:
                args['city'] = st.sidebar.text_input('CITY', 'Karlsruhe')
        
        elif action == 'counties':
            if use_predefined:
                args['state'] = st.sidebar.selectbox('FEDERAL STATE', options=FEDERAL_STATES)
            else:
                args['state'] = st.sidebar.text_input('FEDERAL STATE', 'Baden-Württemberg')
        
        elif action == 'energiewende':
            st.markdown("""### Description\nThis API endpoint is a special endpoint, which calls the `counties` and `node` endpoints several times. Only useful for the Data-Challenge *Energiewende*.""")
            if use_predefined:
                BOUNDS = {**CITIES, **{s: s for s in FEDERAL_STATES}}
                args['boundary'] = st.sidebar.selectbox('BOUNDARY', options=list(BOUNDS.keys()), format_func=lambda k: BOUNDS.get(k))
            else:
                args['boundary'] = st.sidebar.text_input('BOUNDARY', 'Karlsruhe', help="Make sure that this is a valid OSM amenity tag.")

        elif action == 'nodes' or action == 'way':
            args['boundary'] = st.sidebar.text_input('BOUNDARY', 'Baden-Württemberg' if action == 'nodes' else 'Karlsruhe', help="Make sure that this is a valid OSM amenity tag.")
            args['attribute'] = st.sidebar.text_input('OSM Tag Attribute', 'generator:source' if action == 'nodes' else 'leisure', help=f"The OSM Tag, which has to exist on the {action}")
            args['value'] = st.sidebar.text_input('OSM Tag Value', 'wind' if action == 'nodes' else 'playground', help=f"Filters for this value on {action} with tag {args['attribute']}")
            st.sidebar.info('You can rename the tag')
            alias = st.sidebar.text_input('TYPE ALIAS', 'wind turbine' if action == 'nodes' else 'playground')
            if alias.strip() != "":
                args['type_alias'] = alias

    elif example == 'netatmo':
        st.info('The Netatmo API interface is highly experimental. Database upload is not yet implemented.')
        args['action'] = st.sidebar.selectbox('Action', options=['get-data', 'metadata', 'parse'])
        if args['action'] != 'parse':
            args['password'] = st.sidebar.text_input('Passphrase', '', help='The passphrase can be obtained from mirko@hydrocode.de')
            if args['password'] == '':
                st.sidebar.warning('This field can\'t be empty. Please enter a passphrase.')
                st.stop()
            if args['action'] == 'get-data':
                start_date = st.date_input('Load data from', dt.now().date() - td(days=7), help="You can only request 1024 data points of given temporal scling (usually 30min).")
                args['start'] = start_date.strftime('%Y%m%d')
            cc_opts = st.sidebar.radio('Set boundary', options=['city', 'bbox'])
            if cc_opts == 'city':
                args['city'] = st.sidebar.selectbox('City', options=['karlsruhe', 'freiburg'], format_func=lambda c: c.capitalize())
            else:
                args['bbox'] = "[%s]" % st.sidebar.text_input('Bounding box', '49.060234,8.541590,48.974132,8.294625', help='The bounding box has to be in the format: lat_ne,lon_ne,lat_sw,lon_sw')
            args['required-data'] = st.sidebar.multiselect('Required data variables', options=['temperature', 'humidity', 'pressure'], default=['temperature'],  help="Omit Netatmo stations, which do not include these variables")
        else:
            uploaded_file = st.sidebar.file_uploader('Load dump file', type='json', help="Load the Netatmo API dump file here.")
            if uploaded_file is not None:
                args['load'] = uploaded_file.name
            args['use_type'] = st.sidebar.selectbox('Use variable type', options=['temperature', 'humidity', 'pressure'])
            args['fmt'] = st.sidebar.selectbox('Output format', options=['latex','markdown','html','json','csv'])
    else:
        is_defnied = False
    
    # build args
    for k, v in args.items():
        cmd += f" --{k.replace('_', '-')}={v}"

    # ARGS are build now - build the API
    api = HydenvExamples(connection=db.unsafe_get_connection, quiet=False)
    runner = getattr(api, example)

    with st.expander('DOCUMENTATION', expanded=False):
        st.markdown('This is what I found in the code:')
        st.help(runner)

    # show the command
    st.markdown('### CLI command\nThe command below is the corresponding CLI command for the specified API call.')
    st.code(cmd, language='bash')

    if not is_defnied:
        st.info('This example dataset is not yet implemented. Try running it using the cli. Maybe the help page will help as well.')
        st.code(cmd + ' --help', language='bash')
        st.stop()

    # start buttions
    but1 = st.button('RUN CLI', key='but1')
    but2 = st.sidebar.button('RUN CLI', key='but2')

    if but1 or but2:
         # create the classes to handle output
        console_output = st.empty()
        def callback(output):
            console_output.code(output, language='bash')
        
        class Console(StringIO):
            backlog = ""
            def write(self, s):
                self.backlog += s
                super(Console, self).write(s)
                callback(self.backlog)
        
        # handle StdOut
        try:
            orginial_stdout = sys.stdout
            sys.stdout = Console()
            with st.spinner('Executing API call...'):
                if example == 'netatmo':
                    output = runner(**args)
                else:
                    output = None
                    runner(**args)
                runner(**args)
        except Exception as e:
            st.warning('Something went wrong: {}'.format(e))
            st.exception(e)
        finally:
            # restore original StdOut
            sys.stdout = orginial_stdout
        
        if output is not None:
            st.markdown('## CLI output')
            if 'fmt' not in args or args['fmt'] == 'json':
                st.json(output)
            elif args['fmt'] == 'markdown':
                st.markdown(output)
            else:
                st.code(output, language=args['fmt'])
        st.success('Finished')
    st.stop()


def home_page(db: HydenvDatabase):
    st.title('Hydenv CLI overview')
    st.markdown('All Hydenv CLI endpoints which are implemented are available here. You find a short explanation with the call signature and a link to the GUI, if available.')

    # install
    with st.expander('HYDENV DATABASE', expanded=False):
        # INSTALL COMMAND
        st.markdown('Can be used to install new database instance, connect exising ones on the same host or remote and to clean up and re-initialize the database.')
        il, ir = st.columns((9, 1))
        il.code('# install new database\npython -m hydenv database install -i', language='bash')
        iopen = ir.button('OPEN IN GUI', key='open_install')
        if iopen:
            st.session_state.page_name = 'install'
            st.experimental_rerun()
        
        # INIT COMMAND
        st.warning('This will delete all data in the database!')
        il2, ir2 = st.columns((9, 1))
        il2.code('# clean and re-init\npython -m hydenv database init --clean', language='bash')
        ir2_run = ir2.button('RUN HERE', key='run_init')
        if ir2_run:
            db.init(clean=True)
            il2.success('Successfully cleaned and re-initialized the database.')

        # TABLE COMMAND
        tl, tr = st.columns((9, 1))
        tl.code('# show all tables\npython -m hydenv database table --list', language='bash')
        tl.code('# inspect a single table\npython -m hydenv database table --name=osm_nodes --fmt=markdown', language='bash')
        topen = tr.button('TABLE TREE', key='open_table')
        if topen:
            table_names = db.table(list=True, fmt='json')
            tree = {}
            bar = st.progress(0)
            
            # load all tables
            for i, table in enumerate(table_names):
                cols = db.table(table['name'], fmt='json')
                tree[table['name']] ={'TYPE': table['type'], 'ATTRIBUTES': cols}

                bar.progress((i + 1) / len(table_names))

            st.session_state.table_tree = tree
            st.experimental_rerun()
    
    # exercises
    with st.expander('EXERCISES', expanded=True):
        st.markdown('One of the main feature of hydenv-database. You can interactively start SQL exercises, which are tested with your **own database copy** and give you feedback.')
        el, er = st.columns((9, 1))
        el.code('# start exercises\npython -m hydenv gui', language='bash')
        eopen = er.button('OPEN IN GUI', key='open_exercises')
        if eopen:
            st.session_state.page_name = 'exercise'
            st.experimental_rerun()
        
        st.markdown('There is also a full-featured web-application, that is currently not developed any further. But it is still available')
        st.code('python -m hydenv exercises gui --legacy', language='bash')
    
    # examples
    st.markdown('## Examples')
    st.markdown(EXAMPLE_INTRO)
    
    # space
    with st.expander('SPACE MISSIONS', expanded=False):
        st.markdown('The space example set is pre-installed with the package. You don\' need to run this manually.')
        exl, exr = st.columns((9, 1))
        exl.code('# start examples\npython -m hydenv examples space', language='bash')
        exl.code('# Install all normalized tables as well.\npython -m hydenv examples space --normalize', language='bash')

    # hobo
    with st.expander('HOBO', expanded=True):
        st.markdown('One of the main features. The CLI can upload your HOBO measurements, metadata and quality controlled data into your database. Data is available from winter Term 2018 on, quality checked data from term 2019.')

        exhl, exhr = st.columns((9, 1))
        exhl.code('# Upload all from 3 terms\npython -m hydenv examples hobo --terms=[WT,19,WT20,WT21]', language='bash')
        exhl.code('# Upload only metadata\n python -m examples hobo --terms=WT20 --only=metadata')

        hobo_open = exhr.button('OPEN IN GUI', key='open_hobo')
        if hobo_open:
            st.session_state.page_name = 'example_hobo'
            st.experimental_rerun()

    # osm
    with st.expander('OpenStreetMap', expanded=False):
        st.markdown('Hydenv includes some parepared OverpassAPI queries to download infrastructure data from OpenStreetMap.org. These are uploaded into the database, using a slightly different schema.')

        osmr, osml  = st.columns((9, 1))
        osmr.code('# Download OSM city districts\npython -m hydenv examples osm city-districts --city=FR', language='bash')

        osm_open = osml.button('OPEN IN GUI', key='open_osm')
        if osm_open:
            st.session_state.page_name = 'example_osm'
            st.experimental_rerun()
    
    # earthquake
    with st.expander('EARTHQUAKES', expanded=False):
        st.markdown('The earthquake dataset can be used for the exercises. It can upload the raw dataset, or add a normalized version of it.')

        # raw command
        eel1, eer1 = st.columns((9, 1))
        eel1.code('# Upload raw earthquake data\npython -m hydenv examples earthquake', language='bash')
        run_raw = eer1.button('RUN HERE', key='run_raw')
        if run_raw:
            api = HydenvExamples(connection=db.unsafe_get_connection)
            api.earthquake(normalize=False)
            st.success('Successfully uploaded raw earthquake data.')
        
        # normalized command
        eel2, eer2 = st.columns((9, 1))
        eel2.code('# Upload normalized earthquake data\npython -m hydenv examples earthquake --normalize', language='bash')
        run_norm = eer2.button('RUN HERE', key='run_norm')
        if run_norm:
            api = HydenvExamples(connection=db.unsafe_get_connection)
            api.earthquake(normalize=True)
            st.success('Successfully uploaded normalized earthquake data.')
    
    # Netatmo
    with st.expander('Netatmo', expanded=False):
        st.markdown('Hydenv includes a API to download Netatmo weather data. You don\'t need a Netatmo account to use the API. A passphrase will be distributed in course, to use the hydrocode account internally.')
        st.warning('The API only allows the download into a file. Database upload is not yet implemented')

        netl, netr = st.columns((9, 1))
        netl.code('# Download Netatmo data dump into local file\npython -m hydenv examples netatmo get-data --password=<password> --bbox=[49.042909,8.486009,48.985286,8.309699] >> ./ka_dump.json', language='bash')
        netl.code('# Parse the dump created above\npython -m hydenv examples netatmo parse --load=./ka_dump.json --use_type=temperature --fmt=markdown >> temperature.md', language='bash')
        run_netatmo = netr.button('OPEN IN GUI', key='run_netatmo')
        if run_netatmo:
            st.session_state.page_name = 'example_netatmo'
            st.experimental_rerun()

    st.stop()


def event(event_name: str, *args, **kwargs):
    # ger the measurementId
    measurementId = st.session_state.get('measurementId')
    if measurementId is None:
        if st.session_state.get('debug', False):
            print('[Tracker] No measurementId found')
        return

    # get the clientId
    clientId = st.session_state.get('clientId')
    if clientId is None:
        clientId = ''.join(choice(ascii_letters) for _ in range(16))
        st.session_state.clientId = clientId
    
    # do the request
    with HTTPRequest() as http:
        tracker = Tracker(measurementId, http)
        tracker.send(event_name, *args, **kwargs)
        if st.session_state.get('debug', False):
            print(f"[Tracker] Event sent: {event_name} - Payload: {' '. join(args)} {' '.join([f'{k}={v}' for k,v in kwargs.items()])}")


def pageview():
    if'last_page' not in st.session_state:
        st.session_state.last_page = 'none'
    
    # get the current and the last page
    current_page = st.session_state.page_name
    last_page = st.session_state.last_page

    if current_page != last_page:
        # send the pageview
        event('pageview', path=f"/{current_page}", last_page=last_page, title=current_page.capitalize())
        st.session_state.last_page = current_page


def main_app(connection=None, measurementId="G-RLF2LDRQSR", debug=False):
    st.set_page_config(page_title="Excercises by hydrocode", layout="wide")
    
    # hydrocode branding
    l, r = st.sidebar.columns((1,4))
    l.image('https://firebasestorage.googleapis.com/v0/b/hydrocode-website.appspot.com/o/public%2Flogo.png?alt=media&token=8dda885c-0a7d-4d66-b5f6-072ddabf3b02', use_column_width=True)
    r.title('hydrocode powered')
    # set the measurementId
    if 'measurementId' not in st.session_state:
        st.session_state.measurementId = measurementId
    
    if 'debug' not in st.session_state:
        st.session_state.debug = debug

    # check if there is need for install
    db = get_db_con()

    # checkout the app page
    page_name = st.session_state.get('page_name')
    if page_name is None:
        st.session_state.page_name = 'home'
        page_name = 'home'
    # pageview event
    pageview()

    # back button
    if st.session_state.get('connection', False) and page_name != 'home':
        back = st.sidebar.button('Back GUI overview')
        if back:
            st.session_state.page_name = 'home'
            st.experimental_rerun()
    
    # check if a table tree was loaded
    if 'table_tree' in st.session_state:
        st.sidebar.markdown('### DATABASE TABLES')
        st.sidebar.json(st.session_state.table_tree)

    if page_name == 'home':
        home_page(db)
    elif page_name == 'exercise':
        # load the list of solved exercises
        load_solved_list()
        
        # get the tracks and select the exercise
        tracks = get_base_data()
        exercise = manage_session(tracks)

        exercise_page(tracks, db)
    elif page_name == 'install':
        install_page()
    
    elif page_name.startswith('example_'):
        example_page(db)


if __name__ == "__main__":
    import fire
    fire.Fire(main_app)



import os
import sys
import json
from tabulate import tabulate

from stdiomask import getpass
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from sqlalchemy.schema import DropTable
from sqlalchemy.ext.compiler import compiles

from hydenv.models import Base, Term, Variable, Quality, Sensor
from hydenv.examples.space import HydenvSpaceExamples
from hydenv.examples.examples import HydenvExamples
from hydenv.util import env


# Overwriting the sqlalchemy's PostgreSQL compiler to run all DROP statements as DROP ... CASCADE by default.
@compiles(DropTable, 'postgresql')
def _drop_tables_cascade(element, compiler, **kwargs):
    return compiler.visit_drop_table(element) + ' CASCADE'

class HydenvDatabase:
    """
    Hydenv database tool\n
    You can use this tool to install and populate new instances of
    the hydenv database. Works for local, remote and cloud instances.
    This is the admin tool that needs database creation privileges and 
    is usually only run once.
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>
    """
    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}"):
        self.__connection = connection
        self._engine = None
        self._session = None

    @property
    def engine(self):
        if self._engine is None:
            self.__connection = env.build_connection(connection=self.__connection)
            self._engine = create_engine(self.__connection)
        return self._engine

    @property
    def session(self):
        if self._session is None:
            Session = sessionmaker(self.engine)
            self._session = Session()   
        return self._session

    @property
    def unsafe_get_connection(self):
        return self.__connection

    def save(self, user='hydenv', password='hydenv', host='localhost', port='5432', dbname='hydenv'):
        """
        Store the connection information to a file.\n
        Be careful, this function will strore the password to a file in home directory.
        :param usr: PostgreSQL user role
        :param pw: User role password
        :param host: PostgreSQL databse host
        :param port: port of the PostgreSQL daemon
        :param dbname: Database name for Hydenv
        """
        env.store_file(usr=user, pw=password, host=host, port=port, dbname=dbname)

    def connections(self):
        """
        Shows the stored connection information on this machine. 
        Note, this will contain the cleartext password if set.
        """
        return env.read_file()

    def install(self, i=False, db_name='hydenv', user='hydenv', password='hydenv', skip_init=False):
        r"""
        Install the database\n
        You can run the installation interactively by just passing the -i
        flag, **this will ignore all other flags**.
        The connection passed needs to have granted privileges to create
        new database instances. Please note that the Postgis library 
        has to be installed.\n
        The tool will export the connection info to the current command line 
        session. This behavior can be overwritten.
        This behavior 
        :param i: Interactive installation - this ignores all other parameter
        :param db_name: database name
        :param role: The user for the new database instance
        :param password: The users password, HAS TO BE SET
        :param skip_init: Skip init step to run with other user
        :param env: Either 'suppress', 'export' or 'file'\n
            - suppress will do nothing
            - export will expose connection params as environement variables
            - file will write them to an .env file.
        :raises AttributeError: if the password was not given
        """
        if i:
            self.install_interactive()
        else:
            self.install_silent(db_name=db_name, user=user, password=password, skip_init=skip_init)

    def install_interactive(self):        
        ans = ''
        while ans.lower() not in ('y', 'n', 'yes', 'no'):
            print("Use the default PostgreSQL server settings?\n[y]    Yes (localhost:5432)\n[n]    No (changed settings)")
            ans = input("Select [y,n]: ")
        
        if ans == 'y' or ans == 'yes':
            self.__connection = 'postgresql://postgres:{pw}@localhost:5432/postgres'
        elif ans == 'n' or ans == 'no':
            print('Please enter the settings you changed during installation (defaults):')
            host = input('[Postgres] host         (localhost): ') or 'localhost'
            port = input('[Postgres] port              (5432): ') or '5432'
            usr =  input('[Postgres] *superuser*   (postgres): ') or 'postgres'
            self.__connection = 'postgresql://%s:{pw}@%s:%s/postgres' % (usr, host, port)

        print('Now you need to specify the **NEW** database and user:')
        hyd_db =   input('[Hydenv] database name (hyenv): ') or 'hydenv'
        hyd_usr =  input('[Hydenv] username     (hydenv): ') or 'hydenv'
        hyd_pw = getpass('[Hydenv] password: ')

        ans = ''
        while ans.lower() not in ('y', 'n', 'yes', 'no'):
            print("Do you want to initialize the database?\n[y]    Yes - Recommended\n[n]    No")
            ans = input("Select [y,n]: ")
        skip_init = ans == 1
        
        self.install_silent(db_name=hyd_db, user=hyd_usr, password=hyd_pw, skip_init=skip_init)

    def install_silent(self, db_name='hydenv', user='hydenv', password='hydenv', skip_init=False):
        """
        Install the database\n
        The connection passed needs to have granted privileges to create
        new database instances. Please note that the Postgis library 
        has to be installed.\n
        The tool will export the connection info to the current command line 
        session. This behavior can be overwritten.
        This behavior 
        :param db_name: database name
        :param role: The user for the new database instance
        :param password: The users password, HAS TO BE SET
        :param skip_init: Skip init step to run with other user
        :param env: Either 'suppress', 'export' or 'file'\n
            - suppress will do nothing
            - export will expose connection params as environement variables
            - file will write them to an .env file.
        :raises AttributeError: if the password was not given
        """
        # check if the user exists
        with self.engine.connect() as con:
            res = con.execute("SELECT true FROM pg_roles WHERE rolname='{usr}'".format(usr=user)).scalar()
            if res is not True:
                if password == 'changeme':
                    raise AttributeError('Please specify a password like --password=pw')
                con.execute("CREATE USER {usr} WITH PASSWORD '{pw}'".format(usr=user, pw=password))
        
        # create the SQL and commit
        with self.engine.connect() as con:
            con.execute('commit')
            con.execute("CREATE DATABASE %s ENCODING 'UTF8' TEMPLATE 'template0';" % db_name)

        # grant 
        with self.engine.connect() as con:
            con.execute('commit;')
            # grant all on the database
            con.execute('GRANT ALL PRIVILEGES ON DATABASE "{db}" TO {usr};'.format(db=db_name, usr=user))

        # build the connection to the new database
        chunks = self.__connection.split('/')
        uri = ''.join([chunks[0], '//', chunks[2], '/', db_name])
        engine = create_engine(uri)

        # since postgres 15, we need to grant to public as well
        with engine.connect() as con:
            con.execute(f'GRANT ALL ON SCHEMA public TO {user};')
            con.execute('commit;')
        
        # connect and install postgis
        with engine.connect() as con:
            con.execute('CREATE EXTENSION postgis;')
            res = con.execute('Select PostGis_Full_Version();').scalar()
            
        # rebuild the connection
        host_port = self.__connection.split('@')[1].split('/')[0]
        host, port = host_port.split(':')
        self.__init__('postgresql://{usr}:{pw}@{host}:{port}/{db}'.format(
            usr=user, pw=password, host=host,port=port, db=db_name
        ))
        # expose conn if 
        self.__expose_con('file', usr=user, pw=password, host=host, port=port, dbname=db_name)
        
        if skip_init:
            return res
        else:
            print(res)
            self.init(clean=True)

    def init(self, clean=False):
        """
        Create all tables\n
        It is highly recommended to run this command with a user connected 
        who has less privileges than postgres.
        :param clean: If True, any existing table will be dropped
            before creating the tables.
        """
        # drop if existing
        if clean:
            # drop the models
            Base.metadata.drop_all(bind=self.engine)

            # drop other stuff
            cli = HydenvExamples(connection=self.__connection, quiet=True)
            cli.clean()

        
        # creata tables
        Base.metadata.create_all(bind=self.engine)

        # import default Terms
        if clean:
            Term.defaults(session=self.session)
            Variable.defaults(session=self.session)
            Quality.defaults(session=self.session)
            Sensor.defaults(session=self.session)
            cli = HydenvSpaceExamples(connection=self.__connection)
            cli.run(quiet=True)
        
        print('Done')

    def execute(self, sql: str, safe=True, json=False):
        """
        Execute any SQL query.\n
        Note: Do not expose this function in untrusted environments.
        :param sql: SQL query to run
        :param safe: If True (default), only a single SELECT is allowed
        :param json: If given, a list of dicts is returned, containing all column names
        """
        # get rid of comments
        sql = '\n'.join([line for line in sql.split('\n') if not line.strip().startswith('--')])
        
        if safe:
            if not sql.lower().strip().startswith('select') and not sql.lower().strip().startswith('with'):
                raise AttributeError('[SAFE MODE] only SELECT queries are allowed.')
            sql = sql.split(';')[0]
        
        with self.engine.connect() as con:
            result = con.execute(sql)
        
        if json:
            return [{k:v for k,v in row.items()} for row in result]
        else:
            return result.fetchall()
        
    def explain(self, sql: str, path=None, fmt='json', full=False, suppress_rollback=False) -> dict:
        """
        Analyse an sql query.\n
        You can pass any SQL query to run a perf test on the query planner.
        The output will be a dictionary containing the Planning and Execution 
        time for the whole query. The key 'Short' will contain a nested 
        brief summary of all steps' execution time and cost.
        If full is set to True, The raw query Planner ANALYSE output will be 
        attached as 'Full' key to the dictionary.
        :param sql: SQL query to analyse.
        :param path: If a file path is given, the output will be saved to that file
        :param fmt: One of ['json', 'text', 'xml', 'yaml']. Text is only suitable for command
            line output. XML and Yaml are suitable for file output. JSON can be used for 
            all outputs and is most sophisticated.
        :param full: If set to True, The plain ANALYSE output will be added to output.
            Only works with JSON output.
        """
        # check if sql is a file:
        if os.path.exists(sql):
            with open(sql, 'r') as f:
                sql = f.read()
        
        # build analyse query
        if not sql.lower().startswith('explain'):
            sql = "EXPLAIN (ANALYSE, FORMAT %s) %s;" % (fmt, sql)
        
        
        # execute the query
        with self.engine.connect() as con:
            # begin transaction if needed
            if not suppress_rollback:
                con.execute('BEGIN;')
            
            # run the query
            result = con.execute(sql)
            d = result.fetchall()

            # rollback transaction if needed
            if not suppress_rollback:
                con.execute('ROLLBACK;')
                
        # handle other output than JSON
        if fmt.lower() in ('xml', 'yaml'):
            return d[0][0]
        elif fmt.lower() == 'text':
            return '\n'.join([_[0] for _ in d])
        
        # extract info 
        while not isinstance(d, dict) or len(d) == 1:
            d = d[0]

        # build the query plan object
        explain = dict(
            Planning=d.get('Planning Time'), 
            Execution=d.get('Execution Time')
        )
        plan = d['Plan']

        # TODO maybe I need this somewhere else -> move into util?
        def parse_node(nodes):
            if isinstance(nodes, dict):
                newnode = {'action': nodes['Node Type']}
                if 'Actual Total Time' in nodes and 'Actual Startup Time' in nodes:
                    newnode['time'] = nodes['Actual Total Time'] - nodes['Actual Startup Time']
                if 'Total Cost' in nodes and 'Startup Cost' in nodes:
                    newnode['cost'] = nodes['Total Cost'] - nodes['Startup Cost']

                if 'Plans' in nodes:
                    newnode['plans'] = parse_node(nodes['Plans'])
                return newnode
            elif isinstance(nodes, list):
                return [parse_node(node) for node in nodes]
            else:
                return nodes
        
        # parse all nodes
        explain['Short'] = parse_node(plan)
        if full:
            explain['Full'] = plan
        
        if path is None:
            return explain
        else:
            with open(path, 'w') as f:
                json.dump(explain, f, indent=4)    

    def table(self, name: str = None, list: bool = False, fmt: str = 'github'):
        """
        List all tables in the database.\n
        You can either pass a name of a table to get the structure of
        that table or pass the --list flag to list all tables.
        :param name: Name of the table to get the structure of
        :param list: If True, list all tables
        :param fmt: One of ['json', 'markdown', 'github']. Output format.
        """
        if name is None and not list:
            print('[ERROR] You need to specify a table name or use the --list flag.')
            return
        
        # build the query
        if list:
            sql = "SELECT table_name as name, table_type as type FROM information_schema.tables WHERE table_schema = 'public';"
        else:
            sql = f"SELECT column_name as name, data_type as type, is_nullable as nullable, column_default as default, ordinal_position FROM information_schema.columns WHERE table_name = '{name}' ORDER BY ordinal_position;"   

        # reun
        result = self.execute(sql, json=True)

        if fmt.lower() == 'json':
            return result
        else:
            return tabulate(tabular_data=result, headers='keys', tablefmt=fmt)

    def __expose_con(self, action, **kwargs):
        if action.lower() == 'suppress':
            return
        elif action.lower() == 'export':
            env.expose(**kwargs)
        elif action.lower() == 'file':
            env.store_file(**kwargs)
        else:
            raise AttributeError("env has to be one of 'suppress', 'export' or 'file'")


if __name__=='__main__':
    import fire
    fire.Fire(HydenvDatabase)



import pandas as pd
import os
import ntpath
import re
import warnings
from string import ascii_lowercase
from random import choice
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from datetime import datetime as dt
from datetime import timedelta as td
from progressbar import ProgressBar

from hydenv.util import env
from hydenv.measurements import HydenvMeasurements
from hydenv.models import Metadata, Term


def read_file(fname, txtfmt=False, dayfirst=False):
#	if txtfmt:
	if fname.lower().endswith('.txt'):
		# HOBO TXT format
		df = pd.read_csv(fname, skiprows=2, header=None, sep=r'\s+', thousands=',', na_values='Logged')
		try:
			df['tstamp'] = [dt.strptime(row[1] + ' ' + row[2], '%d-%m-%y %H:%M:%S') for i, row in df.iterrows()]
		except Exception as e:
			try:
				df['tstamp'] = pd.to_datetime([row[1] + ' ' + row[2] for i, row in df.iterrows()], infer_datetime_format=True)
			except:
				raise e
		df.drop([0,1,2], axis=1, inplace=True)
		df.columns = ['temperature', 'light', 'tstamp']
#	else:
	elif fname.lower().endswith('.csv'):
		# HOBO CSV format
		df = pd.read_csv(fname, skiprows=2, header=None, parse_dates=[1], usecols=[1,2,3], dayfirst=dayfirst)
		#df.drop(0, axis=1, inplace=True)
		df.dropna(inplace=True)
		df.columns = ['tstamp', 'temperature', 'light']
	else:
		raise ValueError('Only csv and txt raw data files supported right now.')

	# make sure temperature and light are floats
	df.temperature = df.temperature.astype(float)
	df.light = df.light.astype(float)

	# drop na
	df.dropna(how='any', axis=0, inplace=True)
	
	return df


def read_hourly_file(fname):
	if fname.endswith('.tsv'):
		# read the file
		df = pd.read_csv(fname, sep=r'\s+')

		# since 2020 'date' is renamed to 'dttm'
		df.rename(columns={'dttm': 'date'}, inplace=True)

		# create the hourly datetime index
		conv = lambda r: dt.strptime(r[0], '%Y-%m-%d') + td(hours=int(r[1]))
		idx = df.apply(conv, axis=1)

		# set index
		df['tstamp'] = idx

		# drop old index
		df.drop(['date', 'hour'], axis=1, inplace=True)
		df.columns = ['value', 'origin', 'tstamp']
	elif fname.endswith('.csv'):
		# read the file
		df = pd.read_csv(fname)

		# rename the columns
		df.columns = [s.lower() for s in df.columns]
		df.rename(columns={'th': 'value', 'date': 'tstamp', 'dttm': 'tstamp', 'temp': 'value'}, inplace=True)

		# convert datetime
		df['tstamp'] = [dt.strptime(_, '%Y-%m-%d %H:%M:%S') for _ in df.tstamp]
	else:
		raise AttributeError('The file format is not yet supported.')
	return df


class HydenvHoboImporter:
	"""
    Measurements management tool\n
    This is the main tool for CRUD operations on the database.
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>

    """
	def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}"):
        # substitute and save
		self.__connection = env.build_connection(connection=connection)
		self.engine = create_engine(self.__connection)
		Session = sessionmaker(bind=self.engine)
		self.session = Session()

	def metadata(self, url: str, sheet_suffix: str, term: str = None, if_exists='append', quiet=True, dry=False):
		"""
		Copy metadata from google tables.\n
		Make sure that the URL format is correct. Open the correct sheet and copy the url,
		then change the last parameter from 'edit#gid=...' to '/export?gid=...'
		:param url: URL to the table
		:param sheet_suffix: Suffix to identify the correct sheet in the respective source
		:param term: the short name of the term, like WS17
		:param if_exists: How to handle existing table. Can be one of:
			- append (default) - append data to the table
			- fail - raise an error and exit
			- replace - drop the table and recreate
		"""
		# switch the type of URL given:
		if 'google.com' in url:
			warnings.warn("Using the google docs is deprecated and only supported until WT22. Please use the BWSync&Share doc.", category=DeprecationWarning)
			url += sheet_suffix
			# use correct URL
			if 'edit#gid=' in url:
				url = url.replace('edit#', 'export?')
			if not 'format=csv' in url:
				url += '&format=csv'

			# download and make a copy
			df = pd.read_csv(url, skiprows=1)
		
		# bwsyncandshare
		elif 'bwsyncandshare' in url:
			df = pd.read_excel(url, skiprows=1, sheet_name=sheet_suffix)
		
		# check for local path
		elif os.path.exists(url):
			df = pd.read_excel(url, skiprows=1, sheet_name=sheet_suffix)
		
		else:
			raise ValueError("The given url is not of known format.")
		
		# make a copy of the downloaded/loaded source
		orig = df.copy()

		# remove clear names
		if 'name' in df.columns:
			df.drop('name', axis=1, inplace=True)
		
		# remove anything without device id
		df.rename({'hobo_id': 'device_id'}, axis=1, inplace=True)
		df = df.where(~df.device_id.isnull()).dropna(how='all')
		# df['device_id'] = df.device_id.astype(int)
		df['device_id'] = ['%d' % int(_) for _ in df.device_id]

		# convert lon lat
		df = df.where(~df.longitude.isnull()).dropna(how='all')
		df = df.where(~df.latitude.isnull()).dropna(how='all')
		df['location'] = df[['longitude', 'latitude']].apply(lambda r: 'SRID=4326;POINT (%s %s)' % (r[0], r[1]), axis=1)
		df.drop(['longitude', 'latitude'], axis=1, inplace=True)
		
		# drop any empty row
		df.dropna(axis=1, how='all', inplace=True)
		
		# print the dropped entries
		if not quiet:
			if len(df) != len(orig):
				print('Some entries are missing metadata:\n----------------------------------')
				print(orig.iloc[[i for i in orig.index if i not in df.index],].to_string())


		# check if the sensor 'hobo' exists
		cli = HydenvMeasurements(self.__connection)
		hobo = cli.read('Sensor', return_query=True, name='hobo').first()
		semester = cli.read('Term', return_query=True, short=term).first()
		
		# if there is no HOBO, create one
		if hobo is None:
			hobo = cli.create('Sensor', name='hobo')
		
		# if semester is given, add
		if semester is not None:
			df['term_id'] = semester.id
		
		# add sensor info
		df['sensor_id'] = hobo.id

		# check if the hobo ids already exist
		check_sql = 'SELECT device_id FROM metadata WHERE %s device_id IN (%s)' % (
			'term_id=%d AND ' % semester.id if semester.id is not None else '',
			','.join(["'%s'" % str(_) for _ in df.device_id.values])
			)
		with self.engine.connect() as con:
			available_ids = [_[0] for _ in con.execute(check_sql)]
			# filter df
			df = df.where(~df.device_id.isin(available_ids)).dropna()
			if df.empty:
				print('[Warning]: All HOBO ids are already in the database. Remove them first for re-uploading.')
				return

		# on dry run - return or print
		if dry:
			if not quiet:
				return df.to_markdown()
			else:
				return df
		
		# upload
		for d in df.to_dict('records'):
			try:
				self.session.add(Metadata(session=self.session, **d))
				self.session.commit()
			except Exception as e:
				self.session.rollback()
				if quiet:
					raise e
				else:
					print('[ERROR]: %s' % str(e))

	def upload_raw_data(self, filename: str, meta_id: int, variable=None, dry=False):
		"""
		Upload an Hobo file to the database. 
		:param filename: Path to the file for upload
		:param meta_id: Database ID of your device
		"""
		# check which variable should be uploaded
		if variable is None:
			self.upload_raw_data(filename, meta_id, 'temperature')
			self.upload_raw_data(filename, meta_id, 'light')
			return
		
		# figure out the term
		term_name = self.session.query(Term.short).join(Metadata).filter(Metadata.id==meta_id).scalar()
		dayfirst = term_name in ('WT21', 'WT22')

		# load the file
		try:
			data = read_file(filename, dayfirst=dayfirst)
		except Exception as e:
			print("Parsing file '%s' was not successfull.\nDo not edit the files by hand!\nError: %s " % (filename, str(e)))
			return

		# search the metadata
		cli = HydenvMeasurements(connection=self.__connection)

		# get the ovariable ids
		var_t = cli.read('Variable', name=variable, return_query=True).one()

		# upload temperature
		temp = data[['tstamp', variable]].copy()
		temp.columns = ['tstamp', 'value']
		temp['meta_id'] = meta_id
		temp['variable_id'] = var_t.id

		if dry:
			return temp

		# create database engine
		engine = create_engine(self.__connection)
		
		# create a temporary table name
		name = ''.join([choice(ascii_lowercase) for _ in range(16)])

		# upload to temporary table
		temp.to_sql(name, engine, index=False)
		engine.execute('INSERT INTO raw_data (meta_id, variable_id, tstamp, value) SELECT meta_id, variable_id, tstamp, value FROM %s ON CONFLICT (meta_id, variable_id, tstamp) DO NOTHING' % name)
		engine.execute('UPDATE raw_data r SET value=t.value FROM %s t WHERE r.meta_id=t.meta_id AND r.tstamp=t.tstamp AND r.variable_id=t.variable_id' % name)

		# delete the temp table
		engine.execute("DROP TABLE %s" % name)

		# return the number of rows
		return len(temp)

	def upload_q_data(self, filename: str, meta_id: int):
		"""
		Upload an Hobo file to the database. 
		:param filename: Path to the file for upload
		:param meta_id: Database ID of your device
		"""
		
		# load the file
		try:
			data = read_hourly_file(filename)
		except Exception as e:
			print("Parsing file '%s' was not successfull.\nDo not edit the files by hand!\nError: %s " % (filename, str(e)))
			return

		# search the metadata
		cli = HydenvMeasurements(connection=self.__connection)

		# get the ovariable ids
		var_t = cli.read('Variable', name='temperature', return_query=True).one()
		h_flag = cli.read('Quality', short='H', return_query=True).one()
		r_flag = cli.read('Quality', short='R', return_query=True).one()
		flag_map = {'H': h_flag.id, 'R': r_flag.id}

		# upload temperature
		temp = data[['tstamp', 'value']].copy()
		temp['quality_flag_id'] = data.origin.replace(flag_map)
		temp['meta_id'] = meta_id
		temp['variable_id'] = var_t.id

		# create database engine
		engine = create_engine(self.__connection)
		
		# create a temporary table name
		name = ''.join([choice(ascii_lowercase) for _ in range(16)])

		# upload to temporary table
		temp.to_sql(name, engine, index=False)
		try:
			engine.execute('INSERT INTO data (meta_id, variable_id, tstamp, value, quality_flag_id) SELECT meta_id, variable_id, tstamp, value, quality_flag_id FROM %s ON CONFLICT (meta_id, variable_id, tstamp) DO NOTHING' % name)
			engine.execute('UPDATE data d SET value=d.value, quality_flag_id=d.quality_flag_id FROM %s t WHERE d.meta_id=t.meta_id AND d.tstamp=t.tstamp AND d.variable_id=t.variable_id' % name)
		except Exception as e:
			print("[ERROR]: File: '%s' errored. Message: %s" % (filename, str(e)))
		# delete the temp table
		engine.execute("DROP TABLE %s" % name)

		# return the number of rows
		return len(temp)

	def folder(self, path='.', match=r'[0-9]+\.(txt|csv)', is_quality=False, term: str = None, quiet=True):
		"""
		Upload the whole folder content given at path.\n
		Path defaults to the current location, but can be replaced.
		The default match pattern is '[0-9]+\.(txt|csv)'.
		:param path: Location to search, defaults to  './'
		:param match: Regular expression to match all data files
		:param term: If applicable, add the term for the data like: WS17
		"""
		flist = [os.path.join(path, f) for f in os.listdir(path) if re.match(match, f)]
		
		# load the term if any
		cli = HydenvMeasurements(self.__connection)
		if term is not None:
			semester = cli.read('Term', return_query=True, short=term).first()
			if semester is None:
				raise AttributeError("Term '%s' not found." % term)
			else: 
				term_id = semester.id
		else: 
			term_id = None

		# build a progressbar
		if not quiet:
			bar = ProgressBar(max_value=len(flist), redirect_stdout=True)
		
		# load all files
		for i, fname in enumerate(flist):
			try:
				if is_quality:
					device_id = ntpath.split(fname)[-1].split('.')[0].strip('_-Tht')
				else:
					device_id = ntpath.split(fname)[-1].split('.')[0]
				meta = cli.read('Metadata', return_query=True, device_id=device_id, term_id=term_id).first()
			except:
				print("File '%s' cannot be processed." % fname)
				continue
			
			# if there is no metadata, skip
			if meta is None:
				print('File %s references HOBO ID=%s, which is not found.' % (fname, device_id))
			else:
				try:
					# upload
					if is_quality:
						self.upload_q_data(filename=fname, meta_id=meta.id)
					else:
						self.upload_raw_data(filename=fname, meta_id=meta.id)
				except Exception as e:
					print(f"[ERROR]: File: {fname} errored. Message: {str(e)}")
			if not quiet:
				bar.update(i + 1)			


class HydenvHoboImporterCli(HydenvHoboImporter):
	def upload(self, filename: str, meta_id: int, variable=None):
		# run the function
		count = super(HydenvHoboImporterCli, self).upload(filename=filename, meta_id=meta_id, variable=variable)
		if count is not None:
			print('[%s]: Uploaded %d %s measurements.' % (meta_id, count, variable))

	def folder(self, path='.', match=r'[0-9]+\.(txt|csv)', term: str = None):
		super(HydenvHoboImporterCli, self).folder(path=path, match=match, term=term, quiet=False)


if __name__=='__main__':
	import fire
	fire.Fire(HydenvHoboImporterCli)



from functools import wraps
import json

from tabulate import tabulate

from hydenv import models

def check_overrides(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        self = args[0]
        # check if skip_overrides is set
        if not kwargs.get('skip_overrides', False):
            model_name = args[1]
            override_name = '%s_%s' % (func.__name__, model_name.lower())
            if hasattr(self, override_name):
                print(kwargs)
                ov_func = getattr(self, override_name)
                return ov_func(**{k:v for k,v in kwargs.items() if k != 'skip_overrides'})
        
        # if it hasn't return, run the core func
        return func(*args, **{k:v for k,v in kwargs.items() if k != 'skip_overrides'})
    return wrapper


def stringify_output(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        self = args[0]
        if hasattr(self, 'fmt'):
            fmt = self.fmt
        else:
            fmt = 'json'
        
        # run the function
        output = func(*args, **kwargs)

        if not isinstance(output, (tuple, list)):
            output = [output]

        # handle output
        out = []
        for o in output:
            if isinstance(o, models.Base):
                d = dict()
                for col in o.__table__.columns:
                    d[col.name] = str(getattr(o, col.name))
                out.append(d)
            else:
                out.append(str(o))
        
        if fmt == 'print':
            return tabulate(out, headers='keys', tablefmt='psql')
        elif fmt == 'json':
            return json.dumps(out, indent=4)
        elif fmt in ('html', 'rst', 'latex'):
            return tabulate(out, headers='keys', tablefmt=fmt)
        else:
            return out
    return wrapper



"""
Examples
--------
This CLI loads example files from Github (or other sources) 
and loads them into the given database. 

Note
----
The database has to be installed and initialized before. 
Do that by hand or use the two CLI commands:

.. code-block:: bash

    python -m hydenv database install  --connection=postgresql://postgres:<adminpassword>@localhost:5432/postgres
    python -m hydenv database init --clean --connection=postgresql://hydenv:hydenv@localhost:5432/hydenv

"""
import progressbar
import os
import shutil
import zipfile
import requests
import io

from hydenv.util import env
from hydenv.scripts.hobo_import import HydenvHoboImporter


class HydenvHoboExamples:
    """
    Hobo Example Loader.\n
    Loads HOBO example data from Github (or other sources) and loads them into the
    given database. The database has to be installed and initialized before. 
    Do that by hand or use the two CLI commands:
        python -m hydenv database install  --connection=postgresql://postgres:<adminpassword>@localhost:5432/postgres
        python -m hydenv database init --clean --connection=postgresql://hydenv:hydenv@localhost:5432/hydenv
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>
    """
    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}", metadata_source: str = 'bwsyncandshare'):
        """Hydenv-HOBO CLI\n
        Use this cli to load HOBO metadata and data to the database.
        You can connect local or remote PostgreSQL/PostGIS databases, if the install/init
        command was issued before, to create the tables.
        With metadata_source, the shared metadata document can be loaded from google tables,
        BWSync&Share excel file or from a local excel file.
        :param metadata_source: Can be either 'google', 'bwsyncandshare' for remote or a path for local.
        """
        # substitute and save
        self.__connection = env.build_connection(connection=connection)

        # set the source
        self._metadata_source = metadata_source

        # set some URLs
        self._hobo_table_google_url = "https://docs.google.com/spreadsheets/d/1uvI1a_OOnCcFjUbHAO1N5nEWejnI5y5vCNt6eueHrOc/edit"
        self._hobo_table_bw_url = "https://bwsyncandshare.kit.edu/s/pXD9LwEp6q3bZPn/download/HOBO_Meta.xlsx"
        self._hobo_data_url = "https://github.com/data-hydenv/data/archive/master.zip"
        # TODO: this can be done better
        self.__all_terms = ['WT17', 'WT18', 'WT19', 'WT20', 'WT21', 'WT22', 'WT23']
        self.__hobo_gid_map = {
            'WT17': "#gid=1594319231",
            'WT18': "#gid=1500046150",
            'WT19': '#gid=651397949',
            'WT20': "#gid=0",
            'WT21': "#gid=878625339",
            'WT22': "#gid=2042487505"
        }
        self.__hobo_sheet_map = {
            'WT17': "Metadata 2017",
            'WT18': "Metadata 2018",
            'WT19': 'Metadata 2019',
            'WT20': "Metadata 2020",
            'WT21': "Metadata 2021",
            'WT22': "Metadata 2022",
            'WT23': "Metadata 2023"
        }
        self.__hobo_data_map = {
            'WT17': 'hobo/2017/',
            'WT18': 'hobo/2018/',
            'WT19': 'hobo/2019/',
            'WT20': 'hobo/2020/',
            'WT21': 'hobo/2021/',
            'WT22': 'hobo/2022/',
            'WT23': 'hobo/2023/'
        }
    
    def run(self, terms='all', data_path='download', only=None, quiet=True, dry=False):
        if only is None or only == 'metadata':
            # Get the metadata
            if not quiet:
                print('Uploading Metdata')
            self._upload_hobo(terms=terms, quiet=quiet)
            if not quiet:
                print('Done.')
        
        # download
        if only is None or only != 'metadata':
            try:
                if data_path == 'download':
                    if not quiet:
                        print('Downloading data repository. This may take a minute.')
                    self._download_data()
                    data_path = './data-master/'
                    if not quiet:
                        print('Done.')
                
                # upload
                if not quiet:
                    print('Start uploading. You can grab a coffee...')
                
                # check data to be uploaded
                if only == 'raw-data' or only == 'raw' or only == 'r':
                    data_f = 'raw'
                elif only == 'quality-data' or only=='quality' or only == 'q':
                    data_f = 'quality'
                else:
                    data_f = 'all'
                self._upload_data(data_path, terms=terms, data=data_f, quiet=quiet)

            except Exception as e:
                raise e
            finally:
                # delete the temp data 
                if os.path.exists(data_path):
                    shutil.rmtree(data_path)
        
        if not quiet:
                print('\nDone.')
        
    def _upload_hobo(self, terms='all', quiet=True, dry=False):
        # if all years, are requested, build the list
        if terms == 'all':
            terms = self.__all_terms
        if not isinstance(terms, list):
            terms = [terms]

        # build the cli
        cli = HydenvHoboImporter(connection=self.__connection)
        
        # build the upload bar
        if not quiet:
            bar = progressbar.ProgressBar(max_value=len(terms), redirect_stdout=True)
        
        # get the supported term key
        supported_term_keys = list(self.__hobo_gid_map.keys()) if self._metadata_source.lower() == 'google' else list(self.__hobo_sheet_map.keys())
        

        # upload Metadata
        for i, term in enumerate(terms):
            # get the supported term keys
            if not term in supported_term_keys:
                print('Term %d has no mapped table in source %s' % (term, self._metadata_source))
                continue

            # build the url
            # set the base url
            if self._metadata_source.lower() == 'google':
                url = self._hobo_table_google_url
                suffix = self.__hobo_gid_map[term]
            elif self._metadata_source.lower() == 'bwsyncandshare':
                url = self._hobo_table_bw_url
                suffix = self.__hobo_sheet_map[term]
            else:
                # assume local
                url = self._metadata_source
                suffix = self.__hobo_sheet_map[term]

            # upload
            cli.metadata(url=url, sheet_suffix=suffix, term=term, quiet=quiet, dry=dry)
            if not quiet:
                bar.update(i + 1)

    def _download_data(self):
        # download the zip into memory
        r = requests.get(self._hobo_data_url)
        z = zipfile.ZipFile(io.BytesIO(r.content))

        # save to current directory
        z.extractall()
    
    def _upload_data(self, path,  terms='all', data='all', quiet=True):
        # if all years, are requested, build the list
        if terms == 'all':
            terms = self.__all_terms
        if not isinstance(terms, list):
            terms = [terms]

        # build the cli
        cli = HydenvHoboImporter(connection=self.__connection)

        for term in terms:
            if data == 'all' or data == 'raw':
                # raw data
                p = os.path.join(path, self.__hobo_data_map[term], 'raw')
                cli.folder(path=p, term=term, is_quality=False, quiet=quiet)

            if data == 'all' or data == 'quality':
                # quality checked data
                p = os.path.join(path, self.__hobo_data_map[term], 'hourly')
                cli.folder(path=p, match=r'[0-9]+_Th\.(tsv|csv)', is_quality=True, term=term, quiet=quiet)
            



import json
import requests

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from shapely.geometry import shape
from progressbar import ProgressBar

from hydenv.util import env
from hydenv import models

class HydenvWorldBorderExample:
    """
    World Border Dataset Example Loader.\n
    Loads world border example data from Github and load it into the
    given database. The database has to be installed and initialized before. 
    Do that by hand or use the two CLI commands:
        python -m hydenv database install  --connection=postgresql://postgres:<adminpassword>@localhost:5432/postgres
        python -m hydenv database init --clean --connection=postgresql://hydenv:hydenv@localhost:5432/hydenv
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>
    :param overwrite_url: string - overwrite the default data url to load other data. 
    :param quiet: bool - suppresses print output to stdout

    """

    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}", overwrite_url=None):
        # substitute and save
        self.__connection = env.build_connection(connection=connection)
        self.engine = create_engine(self.__connection)
        Session = sessionmaker(bind=self.engine)
        self.session = Session()

        # set some URLs
        if overwrite_url is None:
            self._data_url = 'https://github.com/datasets/geo-countries/raw/master/data/countries.geojson'
        else:
            self._data_url = overwrite_url
    
    def run(self, quiet=True):
        """ 
        """
        if not quiet:
            print('Downloading %s...' % self._data_url, end='')

        # download
        response = requests.get(self._data_url)

        if not response.status_code == 200:
            msg = '%d Download failed.\n' % response.status_code
            if not quiet:
                print('error.\n[ERROR]: %s' % msg)
                return
            else:
                raise requests.ConnectionError(msg)
        else:
            if not quiet:
                print('done.')
        
        # parsing data
        if not quiet:
            print('Parsing data...', end='')
        # get the JSON
        wbd = json.loads(response.content)
        features = wbd['features']
        
        # build a dataframe
        if not quiet:
            bar = ProgressBar(max_value=len(features), redirect_stdout=True)
        
        for i, feature in enumerate(features):
            mod = models.WorldBorder(**self._get_record(feature=feature))

            try:
                self.session.add(mod)
                self.session.commit()
            except Exception as e:
                self.session.rollback()
                if not quiet:
                    print('[%d] errrored: %s' % (i + 1, str(e)))
            
            if not quiet:
                bar.update(i + 1)
        
        if not quiet:
            print('\nDone.')

    
    def _get_record(self, feature):
        prop = feature['properties']
        return dict(
            geom='SRID=4326;%s' % shape(feature['geometry']).wkt,
            name=prop['ADMIN'],
            iso_a2=prop['ISO_A2'],
            iso_a3=prop['ISO_A3']
        )



from setuptools import setup, find_packages

def version():
    with open('hydenv/__version__.py', 'r') as f:
        loc = dict()
        exec(f.read(), loc, loc)
        return loc['__version__']


def readme():
    with open('README.md') as f:
        return f.read()


def requirements():
    with open('requirements.txt') as f:
        return f.read().split('\n')


setup(
    name='hydenv',
    version=version(),
    author='Mirko Mälicke',
    author_email='mirko@hydrocode.de',
    description='Database management tools for data lecture',
    long_description=readme(),
    long_description_content_type='text/markdown',
    install_requires=requirements(),
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False
)



import os
import json
from stdiomask import getpass

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, Session

NAME_MAP = dict(
    usr='POSTGRES_USER',
    pw='POSTGRES_PASSWORD',
    host='POSTGRES_HOST',
    port='POSTGRES_PORT',
    dbname='POSTGRES_DBNAME'
)


def store_file(**kwargs):
    fname = os.path.abspath(os.path.join(os.path.expanduser('~'), 'hydenv.env'))
    config = read_file()
    # update
    for k,v in kwargs.items():
        config[NAME_MAP[k]] = v
    
    with open(fname, 'w') as f:
        json.dump(config, f)


def read_file() -> dict:
    fname = os.path.abspath(os.path.join(os.path.expanduser('~'), 'hydenv.env'))
    if os.path.exists(fname):
        with open(fname, 'r') as f:
            config = json.load(f)
    else:
        config = dict()
    return config


def expose(**kwargs):
    for k,v in kwargs.items():
        os.environ[k] = v


def build_connection(connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}") -> str:
    # check if we need to substitute at all
    if '{' not in connection and '}' not in connection:
        return connection

    # check for replacements
    fargs = read_file()
    args = dict()
    if '{usr}' in connection:
        args['usr'] = os.environ.get('POSTGRES_USER', fargs.get('POSTGRES_USER', 'hydenv'))
    if '{host}' in connection:
        args['host'] = os.environ.get('POSTGRES_HOST', fargs.get('POSTGRES_HOST', 'localhost'))
    if '{port}' in connection:
        args['port'] = os.environ.get('POSTGRES_PORT', fargs.get('POSTGRES_PORT', 5432))
    if '{dbname}' in connection:
        args['dbname'] = os.environ.get('POSTGRES_DBNAME', fargs.get('POSTGRES_DBNAME', 'hydenv'))
    if '{pw}' in connection:
        pw = os.environ.get('POSTGRES_PASSWORD', fargs.get('POSTGRES_PASSWORD'))
        if pw is None:
            curr_user = args['usr'] if 'usr' in args else ''
            pw = getpass('PostgreSQL user %s password: ' % curr_user)
        args['pw'] = pw

    # substitute and return
    return connection.format(**args)


def get_session(connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}") -> Session:
    uri = build_connection(connection=connection)

    engine = create_engine(uri)
    SessionCls = sessionmaker(bind=engine)

    return SessionCls()



from typing import List
from progressbar import ProgressBar

from hydenv.examples.hobo import HydenvHoboExamples
from hydenv.examples.space import HydenvSpaceExamples
# from hydenv.examples.customers import HydenvCustomerExamples
from hydenv.examples.osm import HydenvOSMExamples
from hydenv.examples.earthquakes import HydenvEarthquakeExamples
from hydenv.examples.wbd import HydenvWorldBorderExample
from hydenv.examples.gpx import HydenvGPXExample
from hydenv.examples.raspi_logger import HydenvRaspiLoggerExample
from hydenv.examples.owm import HydenvOWMExample
from hydenv.examples.netatmo import HydenvNetatmoExample
from hydenv.examples.camels_de import HydenvCamelsDeExample

CLIs = [HydenvHoboExamples, HydenvSpaceExamples, HydenvOSMExamples, HydenvEarthquakeExamples, HydenvWorldBorderExample, HydenvGPXExample, HydenvRaspiLoggerExample, HydenvOWMExample, HydenvNetatmoExample, HydenvCamelsDeExample]

class HydenvExamples:
    r"""
    Example Loader.\n
    Loads example files from Github (or other sources) and loads them into the
    given database. The database has to be installed and initialized before. 
    Do that by hand or use the two CLI commands:
        python -m hydenv database install  --connection=postgresql://postgres:<adminpassword>@localhost:5432/postgres
        python -m hydenv database init --clean --connection=postgresql://hydenv:hydenv@localhost:5432/hydenv
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>
    """
    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}", quiet=False):
        # substitute and save
        self.__connection = connection
        self.quiet = quiet

    def hobo(self, terms='all', only=None, data_path='download', metadata_source='google'):
        """
        Import HOBO data.\n
        This high level script downloads all neccessary resources for the HOBO 
        measureing campaigns and uploads them to the specified database.
        You can limit the metadata by terms (like WT17, WT18 etc.) and upload only
        'metadata' or 'data'. Default is to upload both.
        :param terms: Either 'all' (default) or a specific term short name (like WT18) to use
        :param data_path: Either 'download' (default) or a local path to the data
        :param only: If set, only the given entities will be uploaded.
        """
        cli = HydenvHoboExamples(connection=self.__connection, metadata_source=metadata_source)
        return cli.run(terms=terms, data_path=data_path, only=only, quiet=self.quiet)

    def space(self, normalize=False):
        """
        Import Space Missions example data.\n
        This high level script downloads a database of all space Mission from 
        1957 until early 2020. The raw, de-normalized dataset is installed along 
        with the hydenv database and used in lecture to introduce SQL and normalization.
        The final normalized structure can be added using the normalize flag.
        :param normalize: If set, the data will be normalized before uplaod.
        """
        cli = HydenvSpaceExamples(connection=self.__connection)
        return cli.run(normalize=normalize, quiet=self.quiet)

    def customers(self, k=500, normalize=False):
        """
        Import Customers example data.\n
        This high level script create a set of k random fake customer purchases 
        in fake stores. The data is not normalized and can be used for the 
        normalization sessions.
        :param k: number of purchases to create
        :param normalize: If True, the normalized scheme will be added as well.
        """
        # cli = HydenvCustomerExamples(connection=self.__connection)
        # return cli.run(quiet=self.quiet, k=k, normalize=normalize)
        return 'The customers CLI has been deprecated.\nA re-implementation is planned for v1.0.0.'

    def customer(self, **kwargs):
        #return self.customers(**kwargs)
        return 'The customers CLI has been deprecated.\nA re-implementation is planned for v1.0.0.'

    def osm(self, action, *args, **kwargs):
        """
        Import OSM example data.\n
        This high level script downloads data from the OSM database and imports
        it into the hydenv database at given connection. You have to specify, 
        which example set should be loaded using the `action` parameter.
        There are the following example sets available:
        
        * energiewende
        * city-districts

        Additionally, you can use the more low level endpoints to access any data.

        * nodes (for POINT geometries)
        * ways (for LINESTRING and POLYGON geometries)

        Using the low level endpoints, you need to specify the `boundary`,
        `attribute` and `value` parameters.
        :param action: The CLI action to perfom. See docs above.
        :param boundary: string - Any valid OSM administrative boundary to limit
            the results to, like `"Baden-Württemberg"` or `"Freiburg im Breisgau"`.
        :param attribute: string - Any valid OSM Tag attribute for filtering.
            In most cases, this will be `amenity`.
        :param value: string - filter value. This is the value that the 
            attribute should have. If value is not given, all nodes with having 
            the tag will be returned. That can be a lot.
        """
        cli = HydenvOSMExamples(connection=self.__connection)

        if action.lower() == 'energiewende':
            return cli.energiewende(quiet=self.quiet, **kwargs)
        elif action.lower() == 'city-districts' or action.lower() == 'city_districts':
            return cli.city_districts(quiet=self.quiet, **kwargs)
        elif action.lower() == 'counties':
            return cli.counties(quiet=self.quiet, **kwargs)
        elif action.lower() == 'node' or action.lower() == 'nodes':
            return cli.run('node', *args, quiet=self.quiet, **kwargs)
        elif action.lower() == 'way' or action.lower() == 'ways':
            return cli.run('way', *args, quiet=self.quiet, **kwargs)
        else:
            print("'%s' is not a supported OSM example." % action)

    def owm(self, save: str = None, fmt='json', variable=['temp', 'humidity'], if_exists='replace', repo_path='extra/weather/data', pattern='*_raw_dump.json',):
        """
        Import OpenWeatherMap example data.\n
        This high level script downloads data from the OpenWeatherMap database and 
        exports the data to the given save path. If you do not specify a save path (default),
        the data will be uploaded to the database.
        :param save: The path to save the data to. If empty, the data will be loaded to the database
        :param fmt: The format of the data. Currently, only json and csv is supported.
        :param variable: The variable to download. Defaults temperature and humidity.
        :param if_exists: Specifies the action if the data already exists. Can be 'replace', 'raise' or 'ignore'.
        :param repo_path: The path to the data *inside* the remote repository. Ususally you don't need to change this.
        :param pattern: The pattern to find the data dumps. Usually you don't need to change this.
        """
        cli = HydenvOWMExample(connection=self.__connection)
        return cli.run(quiet=self.quiet, fmt=fmt, save=save, variable=variable, if_exists=if_exists, repo_path=repo_path, pattern=pattern)

    def netatmo(self, action: str, password: str = None, city: str = None, bbox: List[float] = None, load: str = None, use_type='temperature', fmt='latex', **kwargs):
        """
        Download Netatmo personal weather station data.\n
        This high level script downloads data from the Netatmo API and exports
        it into a raw JSON dump file. The stations can be queried by BBox or a predifined
        city name. Additionally, the API can parse the JSON dump into columnar
        data formats. Upload to the database is not yet supported.
        The supported actions are:

        * get-data: download raw data dump
        * metadata: donload only the station metadata
        * parse: parse a raw data dump

        :param action: One of ['get-data', 'metadata', 'parse']
        :param password: The password to receive a Netatmo token from hydrocode. 
            Ignored for 'action=parse'.
        :param city: The city to download data for. Ignored for 'action=parse'.
        :param bbox: The bounding box to download data for. Ignored for 'action=parse'.
        :param required-data: Filter to query only stations, which include this measure.
            Defaults to 'temperature'. Ignored for 'action=parse'.
        :param save: The path to save the data to. Ignored for 'action=parse'.
        :param load: The filename of the raw data-dump. Only valid for 'action=parse'.
        :param use_type: The Netatmo measure type to extract from the dump.
            Defaults to 'temperature'. Only valid for 'action=parse'.
        :param fmt: The format of the data. Can be one of ['markdown', 'html', 'latex', 'csv'].
            Only valid for 'action=parse'.
        """
        # build the cli
        cli = HydenvNetatmoExample(password=password, quiet=self.quiet)

        # switch the action
        if action.lower() == 'get-data':
            return cli.run(bbox=bbox, city=city, **kwargs)
        elif action.lower() == 'metadata':
            kwargs['metadata_only'] = True
            return cli.run(bbox=bbox, city=city, **kwargs)
        elif action.lower() == 'parse':
            return cli.parse(fname=load, use_type=use_type, fmt=fmt, **kwargs)
        else:
            print('[ERROR] Unknown action: %s' % action)
            return
   
    def earthquake(self, normalize=False):
        """
        Import Earthquakes example data.\n
        This high level script downloads a database of earthquakes from 
        1967 to 2016 with a magnitude > 5.5. This dataset can be used for an advanced
        normalization exercise and as a spatial dataset with global coverage.
        The final normalized structure can be added using the normalize flag.
        :param normalize: If set, the data will be normalized before uplaod.
        """
        cli = HydenvEarthquakeExamples(connection=self.__connection, quiet=self.quiet)
        return cli.run(normalize=normalize)

    def world_borders(self):
        """
        Load a minified version of the World Border dataset.\n
        This dataset should only be used as a overview for maps. There is 
        only very limited information in the tables and the border geometries
        are very coarse.
        """
        cli = HydenvWorldBorderExample(connection=self.__connection)
        return cli.run(quiet=self.quiet)

    def gpx(self, file_name, path_or_url=None):
        """
        Loads a GPX file into the database\n
        This tool can load any GPX file, either from a local path or a remote URL.
        The default path is the URL to the `extra/spatial/activity/` folder in the 
        hydenv data repository on Github.
        :param file-name: File name of the GPX file
        """
        cli = HydenvGPXExample(connection=self.__connection, quiet=self.quiet)
        return cli.run(fname=file_name)

    def activity(self, day='20201217', only=None):
        """
        Activity example.\n
        Loads activity GPX and corresponding raspi-logger raw dumped temperature 
        measurement data to the database.
        :param day: string YYYYMMDD formatted string of the activity day
        """
        gpx_fname = '%s_activity.gpx' % day
        log_fname = '%s_raspi_logger.dump.json' % day
        
        if only is None or only == 'gpx':
            if not self.quiet:
                print('Loading GPX file')
            gpx = HydenvGPXExample(connection=self.__connection, quiet=self.quiet)
            gpx.run(fname=gpx_fname)
        
        if only is None or only == 'raspi' or only == 'dump':
            if not self.quiet:
                print('Loading Raspi dump')
            log = HydenvRaspiLoggerExample(connection=self.__connection)
            log.run(fname=log_fname)
        
        if not self.quiet:
            print('All done.')
    
    def camels_de(self, source_dir: str = None, nuts_filter: str = None, only: str = None, drop: bool = False):
        """
        CAMELS-DE example.\n
        Loads the CAMELS-DE dataset to the database. Currently, only the state API is used,
        as the CAMELS-DE dataset is not yet published. If you want to add data along with
        the metadata (through state API), you need a local copy of the data and set the
        source_dir parameter.
        :param source_dir: string - Path to the CAMELS-DE output_data directory
        :param nuts_filter: string - Filter by federal state using the NUTS level 2 id, like DE1, DE2 ... DEG
        :param only: string - If set, upload only 'metadata' or 'data'
        :param drop: bool - If set, all other parameters are ignored and the CAMELS-DE data is dropped. 
        """
        # initialize the CLI
        cli = HydenvCamelsDeExample(self.__connection, source_dir=source_dir, quiet=self.quiet)
        
        # first check if the tables should be dropped
        if drop:
            if not self.quiet:
                print('Cleaning CAMELS-DE tables...', end='')
            cli.drop() 
            if not self.quiet:
                print('done.')
        else:
            # otherwise run the CLI
            cli.run(only=only, nuts_filter=nuts_filter)

    def clean(self): 
        """
        Cleaning Up.\n
        This command will clean all additional tables that were created by the CLI 
        which do not have a model representation in `hydenv.models`. This includes
        the normalized examples of space and customers.
        Usually, this is only called by the `database --clean` command and a 
        manual invocation is not needed
        """
        if not self.quiet:
            bar = ProgressBar(max_value=len(CLIs), redirect_stdout=True)
        
        # drop all extra tables
        for i, CLI in enumerate(CLIs):
            if hasattr(CLI, 'drop'):
                cli = CLI(connection=self.__connection)
                cli.drop()

                if not self.quiet:
                    bar.update(i + 1)


if __name__=='__main__':
    import fire
    fire.Fire(HydenvExamples)
